import os
import sys
import csv
import time
import json
import random
import orjson
import logging
import tempfile
import typing as t
from copy import copy
from typing import List, Tuple, Optional, Dict, Callable
from collections import defaultdict

import cairosvg
import numpy as np
import networkx as nx
import matplotlib.pyplot as plt
import tensorflow as tf
from kgcnn.data.utils import ragged_tensor_from_nested_numpy
from kgcnn.data.moleculenet import MoleculeNetDataset
from rdkit import Chem
from rdkit.Chem.Draw.rdMolDraw2D import MolDraw2DSVG
from rdkit.Chem import rdDepictor
rdDepictor.SetPreferCoordGen(True)

import graph_attention_student.typing as tc
from graph_attention_student.util import NULL_LOGGER
from graph_attention_student.util import graph_dict_to_list_values
from graph_attention_student.util import update_nested_dict
from graph_attention_student.visualization import draw_extended_colors_graph
from graph_attention_student.visualization import draw_edge_black
from graph_attention_student.visualization import draw_node_color_scatter


class NumericJsonEncoder(json.JSONEncoder):

    def default(self, o: t.Any) -> t.Any:
        if isinstance(o, np.ndarray):
            return o.tolist()
        elif isinstance(o, np.generic):
            return o.item()
        else:
            return super(NumericJsonEncoder, self).default(o)


def numpy_to_native(value):
    if isinstance(value, np.ndarray):
        return value.tolist()
    elif isinstance(value, np.generic):
        return value.item()
    else:
        return value


def process_graph_dataset(dataset: t.List[dict],
                          test_indices: t.Optional[t.List[int]] = None,
                          train_indices: t.Optional[List[int]] = None,
                          use_importances: bool = True,
                          use_graph_attributes: bool = False,
                          ) -> t.Tuple[tuple, tuple, tuple, tuple]:
    """
    Given a list ``dataset`` of GraphDict objects representing the elements of a dataset, this function will
    firstly split these elements according to the given ``train_indices`` and ``test_indices`` and also turn
    this list of graph objects into the appropriate RaggedTensors which can be used to train a
    tensorflow model.
    """
    indices = list(range(len(dataset)))

    # It has to be possible to determine the dataset by either the train or the test indices because now
    # there are cases were we need to be explicitly sure to have an exact number of train elements
    # that has to be divisible by the batch size
    if train_indices is not None and test_indices is None:
        test_indices = [index for index in indices if index not in train_indices]
    elif test_indices is not None and train_indices is None:
        train_indices = [index for index in indices if index not in test_indices]

    labels_train = [dataset[i]['graph_labels'] for i in train_indices]
    labels_test = [dataset[i]['graph_labels'] for i in test_indices]
    nodes_train = [dataset[i]['node_attributes'] for i in train_indices]
    nodes_test = [dataset[i]['node_attributes'] for i in test_indices]
    edges_train = [dataset[i]['edge_attributes'] for i in train_indices]
    edges_test = [dataset[i]['edge_attributes'] for i in test_indices]
    edge_indices_train = [dataset[i]['edge_indices'] for i in train_indices]
    edge_indices_test = [dataset[i]['edge_indices'] for i in test_indices]

    x_train = (
        ragged_tensor_from_nested_numpy(nodes_train),
        ragged_tensor_from_nested_numpy(edges_train),
        ragged_tensor_from_nested_numpy(edge_indices_train)
    )

    x_test = (
        ragged_tensor_from_nested_numpy(nodes_test),
        ragged_tensor_from_nested_numpy(edges_test),
        ragged_tensor_from_nested_numpy(edge_indices_test)
    )

    # Optionally it is also possible to use graph attributes, these are features which describe the graph
    # as a whole rather than its individual elements (nodes / edges). This is however not part of every
    # dataset by default because not all graph neural networks are able to deal with this appropriately.
    # So we also only add this as the fourth element of the X tuple, if the corresponding flag is given
    if use_graph_attributes:
        graph_attributes_train = [dataset[i]['graph_attributes'] for i in train_indices]
        graph_attributes_test = [dataset[i]['graph_attributes'] for i in test_indices]

        x_train = (
            *x_train,
            tf.convert_to_tensor(graph_attributes_train)
        )
        x_test = (
            *x_test,
            tf.convert_to_tensor(graph_attributes_test),
        )

    y_train = (
          np.array(labels_train, dtype=np.single),
    )

    y_test = (
        np.array(labels_test, dtype=np.single),
    )

    # Optionally (although in this project this is the true most of the time) it is possible to load
    # node and edge importances (explanations) as additional training targets from the datasets as well.
    if use_importances:
        node_importances_train = [dataset[i]['node_importances'] for i in train_indices]
        node_importances_test = [dataset[i]['node_importances'] for i in test_indices]
        edge_importances_train = [dataset[i]['edge_importances'] for i in train_indices]
        edge_importances_test = [dataset[i]['edge_importances'] for i in test_indices]

        y_train = (
            *y_train,
            ragged_tensor_from_nested_numpy(node_importances_train),
            ragged_tensor_from_nested_numpy(edge_importances_train)
        )

        y_test = (
            *y_test,
            ragged_tensor_from_nested_numpy(node_importances_test),
            ragged_tensor_from_nested_numpy(edge_importances_test)
        )

    return x_train, y_train, x_test, y_test


def load_eye_tracking_dataset(folder_path: str) -> List[dict]:
    dataset_map = defaultdict(dict)
    for root, dirs, files in os.walk(folder_path):

        for file_name in files:
            name, extension = file_name.split('.')
            file_path = os.path.join(root, file_name)

            if extension in ['png']:
                dataset_map[name]['image_path'] = file_path
                dataset_map[name]['name'] = name

            if extension in ['json']:
                dataset_map[name]['metadata_path'] = file_path
                with open(file_path, mode='r') as json_file:
                    metadata = json.loads(json_file.read())

                # This check fixes a bug, which can happen if there are other JSON files present in the
                # folder and these most likely do not define a similar structure which would cause a
                # KeyError here if we just assumed they would
                if isinstance(metadata, dict) and 'graph' in metadata:
                    dataset_map[name]['metadata'] = metadata
                    # At this point there is already a field "graph" in metadata but the values in that
                    # dict are just lists after being loaded from the JSON file, so we need to convert
                    # them all into numpy arrays (this is actually important!) here
                    dataset_map[name]['graph'] = {key: np.array(value)
                                                  for key, value in metadata['graph'].items()}

        break

    # Now, if we just pass a folder to this function, then there is a real possibility that this folder
    # contains other json or image files as well and we dont want these in our result, which is why we
    # filter out any elements which does not contain both(!) the image and metadata file path fields!
    return [data
            for data in dataset_map.values()
            if 'metadata_path' in data and 'image_path' in data]


def load_eye_tracking_dataset_dict(dataset_path: str,
                                   ratio: float = 1.0,
                                   logger: logging.Logger = NULL_LOGGER,
                                   log_step: int = 1000,
                                   sanitize_dict: bool = False) -> tc.EtDatasetDict:
    """

    *25.10.2022*
    Recently I was confronted with a new dataset that has about half a million entries, and this required
    a serious restructuring of this method due to higher requirements of computational and storage
    efficiency.
    Previously the function worked by iterating the os.listdir() list directly and then checking each
    file's extension and only the correct extensions were added to the dict. This was a more robust
    solution but it required a longer primary loop, which is why I switched to the solution with the
    set of file names now.

    :param dataset_path:
    :param logger:
    :param log_step:
    :param sanitize_dict:
    :return:
    """
    logger.info(f'determining the file names...')
    files = os.listdir(dataset_path)
    names = set()
    for file_name in files:
        name, extension = file_name.split('.')
        if extension in ['png', 'json']:
            png_path = os.path.join(dataset_path, f'{name}.png')
            json_path = os.path.join(dataset_path, f'{name}.json')
            if os.path.exists(png_path) and os.path.exists(json_path):
                names.add(name)

    if ratio < 1.0:
        names = random.sample(names, k=int(len(names) * ratio))

    num_names = len(names)
    num_files = len(files)
    logger.info(f'loading eye tracking dataset from folder {dataset_path} with {num_files} files...')

    start_time = time.time()
    dataset_map = {}
    for c, name in enumerate(names):

        # The dataset canonically represents each element either as png (graph visualization) or
        # json (metadata) if any other files contaminate the folder we do a skip here because afterwards
        # is some code we want to execute for either json or png
        dataset_map[name] = {}
        dataset_map[name]['name'] = name

        # ~ image
        image_path = os.path.join(dataset_path, f'{name}.png')
        dataset_map[name]['image_path'] = image_path

        # ~ metadata
        metadata_path = os.path.join(dataset_path, f'{name}.json')
        dataset_map[name]['metadata_path'] = metadata_path
        dataset_map[name]['index'] = c
        with open(metadata_path, mode='rb') as file:
            content = file.read()
            metadata = orjson.loads(content)
            dataset_map[name]['metadata'] = metadata

        if isinstance(metadata, dict) and 'graph' in metadata:
            dataset_map[name]['metadata'] = metadata
            # At this point there is already a field "graph" in metadata but the values in that
            # dict are just lists after being loaded from the JSON file, so we need to convert
            # them all into numpy arrays (this is actually important!) here
            dataset_map[name]['metadata']['graph'] = {key: np.array(value)
                                                      for key, value in metadata['graph'].items()}

        if c % log_step == 0:
            elapsed_time = time.time() - start_time
            time_per_element = elapsed_time / (c + 1)
            remaining_time = time_per_element * (num_names - c)
            logger.info(f' * loaded ({c}/{num_names})'
                        f' - name: {name}'
                        f' - elapsed time: {elapsed_time:.1f}s ({elapsed_time/3600:.1f}hrs)'
                        f' - remaining time: {remaining_time:.1f}s ({remaining_time/3600:.1f}hrs)'
                        f' - dict overhead: {sys.getsizeof(dataset_map)/1024**2:.2f}MB')

    # It can happen that a dataset is damaged. Maybe it wasn't properly created or maybe there was a problem
    # when copy-pasting or something. It is possible that there is only one of the two required files for
    # an element. In this case there is this optional post-processing which removes such faulty entries
    # from the dict.
    if sanitize_dict:
        logger.info('sanitizing incomplete entries in dataset dict...')
        keys = list(dataset_map.keys())
        for key in keys:
            if 'metadata' not in dataset_map[key] or 'image_path' not in dataset_map[key]:
                del dataset_map[key]

    return dataset_map


# == CHEMISTRY RELATED ======================================================================================

def create_molecule_eye_tracking_dataset(molecule_infos: Dict[str, dict],
                                         dest_path: str,
                                         image_width: int = 900,
                                         image_height: int = 900,
                                         set_attributes_kwargs: dict = {},
                                         logger: logging.Logger = NULL_LOGGER,
                                         log_step: int = 10):
    # First of all, we need to create a temporary folder in which we are going to complete all the
    # intermediate file operations.
    logger.info(f'starting creation of molecule eye tracking dataset with {len(molecule_infos)} elements...')
    with tempfile.TemporaryDirectory() as temp_path:

        logger.info('Writing temporary csv file...')
        # First step is to create a temporary csv file containing the given information about the molecules
        # smiles
        csv_path = os.path.join(temp_path, 'smiles.csv')
        with open(csv_path, mode='w') as csv_file:
            csv_writer = csv.DictWriter(csv_file, fieldnames=['id', 'smiles'], delimiter=',', quotechar='"')
            csv_writer.writeheader()
            for mol_id, mol_info in molecule_infos.items():
                csv_writer.writerow({'id': mol_id, 'smiles': mol_info['smiles']})
            logger.info(f'csv file written to: {csv_path}')

        logger.info('parsing csv file with kgcnn...')
        # With this CSV file we can now instantiate a MoleculeNetDataset instance and process all
        # the molecules
        moleculenet = MoleculeNetDataset(
            file_name=os.path.basename(csv_path),
            data_directory=os.path.dirname(csv_path),
            dataset_name='temp',
            verbose=0
        )
        moleculenet.prepare_data(
            overwrite=True,
            smiles_column_name='smiles',
            add_hydrogen=True,
            make_conformers=False,
            optimize_conformer=False
        )
        moleculenet.read_in_memory(
            label_column_name='smiles',
            add_hydrogen=False,
            has_conformers=False
        )

        # Then we setup the additional callbacks for the property assignment. These functions can be passed
        # to the dataset instance and they are invoked for every molecule object instance in the dataset.
        # That will be the place at which we create the molecule images using RDKit

        dataset_path = dest_path

        def node_coordinates_cb(mg, ds):
            mol = Chem.MolFromSmiles(str(ds['smiles']))
            #mol = mg.mol

            mol_drawer = MolDraw2DSVG(image_width, image_height)
            mol_drawer.SetLineWidth(3)
            mol_drawer.DrawMolecule(mol)
            mol_drawer.FinishDrawing()
            svg_string = mol_drawer.GetDrawingText()
            png_path = os.path.join(dataset_path, str(ds['id']) + '.png')
            cairosvg.svg2png(
                bytestring=svg_string.encode(),
                write_to=png_path,
                output_height=image_height,
                output_width=image_width,
            )

            node_coordinates = []
            for point in [mol_drawer.GetDrawCoords(i) for i, _ in enumerate(mol.GetAtoms())]:
                node_coordinates.append([
                    point.x,
                    image_height - point.y
                ])

            return np.array(node_coordinates)

        def node_indices_cb(mg, ds):
            return np.array(list(range(len(mg.mol.GetAtoms()))))

        kwargs = set_attributes_kwargs
        if 'additional_callbacks' not in kwargs:
            kwargs['additional_callbacks'] = {}

        logger.info('creating custom properties in moleculenet dataset...')
        kwargs['additional_callbacks']['node_coordinates'] = node_coordinates_cb
        kwargs['additional_callbacks']['node_indices'] = node_indices_cb
        kwargs['additional_callbacks']['id'] = lambda mg, ds: str(ds['id'])
        moleculenet.set_attributes(**kwargs)

        # 12.10.2022: Added this section in response to a bug which would cause the program to break if
        # there were some corrupted elements in the original dataset.
        logger.info(f'cleaning moleculenet dataset with {len(moleculenet)} elements...')
        moleculenet.clean(['id', 'node_attributes', 'edge_indices'])
        logger.info(f'dataset length after cleaning: {len(moleculenet)}')

        dataset_length = len(moleculenet)
        logger.info('start writing output files...')
        # Now that the dataset is prepared and even the images are created, the only thing left to do is
        # to create the metadata json file for each element
        for c, g in enumerate(moleculenet):
            mol_id = str(g['id'])
            mol_info = molecule_infos[mol_id]
            json_path = os.path.join(dataset_path, mol_id + '.json')

            # Creating the node_adjacency matrix which is an important part of the GrapDict representation
            length = len(g['node_indices'])
            g['node_adjacency'] = np.zeros(shape=(length, length))
            for i, j in g['edge_indices']:
                g['node_adjacency'][i, j] = 1

            # Adding empty importance tensors here because it will be required that these keys generally
            # exist in further processing steps.
            g['node_importances'] = np.zeros(shape=(g['node_indices'].shape[0], 1))
            g['edge_importances'] = np.zeros(shape=(g['edge_indices'].shape[0], 1))

            metadata = {
                **mol_info,
                'image_width': image_height,
                'image_height': image_height,
                'graph': {k: numpy_to_native(v) for k, v in g.items()}
            }

            with open(json_path, mode='wb') as json_file:
                content = orjson.dumps(metadata)
                json_file.write(content)

            if c % log_step == 0:
                logger.info(f'* ({c}/{dataset_length}) {mol_id}')


def eye_tracking_dataset_from_moleculenet_dataset(moleculenet: MoleculeNetDataset,
                                                  dest_path: str,
                                                  set_attributes_kwargs: t.Dict[str, t.Any],
                                                  smiles_column_name: str = 'smiles',
                                                  clean_keys: t.List[str] = ['edge_attributes'],
                                                  image_width: int = 900,
                                                  image_height: int = 900,
                                                  return_dataset: bool = True,
                                                  logger: logging.Logger = NULL_LOGGER,
                                                  log_step: int = 1000
                                                  ) -> tc.EtDatasetDict:
    dataset_length = len(moleculenet)

    logger.info(f'setting moleculenet attributes...')
    kwargs = {'additional_callbacks': {
        'smiles': lambda mg, ds: str(ds[smiles_column_name])
    }}
    # The way that this function works is that the default options for 'additional_callbacks' are the ones
    # defined above here, but IF the arg "set_attributes_kwargs" also contains an "additional_callbacks"
    # dict, then it is even possible to override the default options we define here, which is probably
    # especially useful for defining custom indexing and naming schemes for the elements.
    kwargs = update_nested_dict(kwargs, set_attributes_kwargs)
    moleculenet.set_attributes(**kwargs)

    logger.info(f'cleaning moleculenet with keys: {clean_keys}')
    moleculenet.clean(clean_keys)

    logger.info(f'generating dataset files...')
    start_time = time.time()
    dataset_map = {}
    for c, d in enumerate(moleculenet):
        index = d['index'] if 'index' in d else c
        name = d['name'] if 'name' in d else str(c)
        smiles = d['smiles']

        # ~ Render the image
        # Technically we are being a bit redundant here:
        mol = Chem.MolFromSmiles(str(smiles))
        # mol = mg.mol

        mol_drawer = MolDraw2DSVG(image_width, image_height)
        mol_drawer.SetLineWidth(3)
        mol_drawer.DrawMolecule(mol)
        mol_drawer.FinishDrawing()
        svg_string = mol_drawer.GetDrawingText()
        image_path = os.path.join(dest_path, name + '.png')
        cairosvg.svg2png(
            bytestring=svg_string.encode(),
            write_to=image_path,
            output_height=image_height,
            output_width=image_width,
        )

        node_coordinates = []
        for point in [mol_drawer.GetDrawCoords(i) for i, _ in enumerate(mol.GetAtoms())]:
            node_coordinates.append([
                point.x,
                image_height - point.y
            ])

        node_coordinates = np.array(node_coordinates)
        d['node_coordinates'] = np.array(node_coordinates)

        # "numpy_to_native" will convert any kind of numpy array to a native datatype such as a list of
        # floats for example. We need to do this to be able to turn it into JSON later.
        graph: tc.GraphDict = {key: numpy_to_native(value)
                               for key, value in d.items()
                               if isinstance(value, np.ndarray) or isinstance(value, np.generic)}

        metadata = {
            'name': name,
            'index': index,
            'image_width': image_width,
            'image_height': image_height,
            'graph': graph
        }
        if 'target' in d:
            metadata['target'] = d['target']

        metadata_path = os.path.join(dest_path, f'{name}.json')
        with open(metadata_path, mode='w') as file:
            content = json.dumps(metadata, cls=NumericJsonEncoder)
            file.write(content)

        if return_dataset:
            dataset_map[name] = {
                'index': index,
                'image_path': image_path,
                'metadata_path': metadata_path,
                'metadata': metadata,
            }

        if c % log_step == 0:
            elapsed_time = time.time() - start_time
            time_per_element = elapsed_time / (c+1)
            remaining_time = time_per_element * (dataset_length - c+1)
            logger.info(f' * created ({c}/{dataset_length})'
                        f' - index: {index} - name: {name}'
                        f' - elapsed time: {elapsed_time:.1f}s ({elapsed_time/3600:.1f}hrs)'
                        f' - remaining time: {remaining_time/3600:.1f}hrs')

    return dataset_map


# == TEXT GRAPHS ============================================================================================

def load_text_graph_dataset(folder_path: str,
                            logger: Optional[logging.Logger] = None,
                            log_interval: int = 25) -> Dict[str, dict]:
    start_time = time.time()
    dataset_map = defaultdict(dict)
    for root, dirs, files in os.walk(folder_path):

        length = len(files)
        for index, file_name in enumerate(files):
            name, extension = file_name.split('.')
            file_path = os.path.join(root, file_name)

            if extension in ['txt']:
                dataset_map[name]['name'] = name
                dataset_map[name]['text_path'] = file_path

            if extension in ['json']:
                dataset_map[name]['metadata_path'] = file_path
                with open(file_path, mode='r') as json_file:
                    metadata = json.loads(json_file.read())

                # This check fixes a bug, which can happen if there are other JSON files present in the
                # folder and these most likely do not define a similar structure which would cause a
                # KeyError here if we just assumed they would
                if isinstance(metadata, dict) and 'graph' in metadata:
                    dataset_map[name]['metadata'] = metadata

                    # We need to convert the graph which is part of the metadata dict such that it's values
                    # are no longer just lists, but numpy arrays instead
                    graph = {k: np.array(v)
                             for k, v in dataset_map[name]['metadata']['graph'].items()
                             if k not in ['node_adjacency']}
                    del dataset_map[name]['metadata']['graph']
                    dataset_map[name]['metadata']['graph'] = graph

            if logger and index % log_interval == 0:
                logger.info(f'({index}/{length}) loaded element {name:<10} - '
                            f'elapsed time: {time.time() - start_time:.2f} seconds')

        break

    # Now, if we just pass a folder to this function, then there is a real possibility that this folder
    # contains other json or image files as well and we dont want these in our result, which is why we
    # filter out any elements which does not contain both(!) the image and metadata file path fields!
    for key in dataset_map.keys():
        data = dataset_map[key]
        if not ('metadata_path' in data and 'text_path' in data):
            del dataset_map[key]

    return dataset_map


# == DATASET GENERATION =====================================================================================

# Secondary graph can be in a subgraph relation towards the primary graph, such that the primary graph is
# allowed to have additional local connections and the results will still be True. That does NOT go the
# other way around though!
def graphs_locally_similar(primary_graph: dict,
                           primary_index: int,
                           secondary_graph: dict,
                           secondary_index: int,
                           consider_edges: bool = False,
                           similarity_func: Callable = lambda a1, a2: np.all(a1 == a2),
                           primary_visited_indices: List[int] = [],
                           secondary_visited_indices: List[int] = []
                           ) -> Tuple[np.ndarray, np.ndarray]:

    if not similarity_func(primary_graph['node_attributes'][primary_index],
                           secondary_graph['node_attributes'][secondary_index]):
        raise StopIteration('The given initial nodes are not similar!')

    primary_visited_edges = [1
                             for i in primary_graph['node_indices']
                             if (primary_graph['node_adjacency'][primary_index, i] or
                                 primary_graph['node_adjacency'][i, primary_index])
                             and i in primary_visited_indices]

    secondary_visited_edges = [1
                               for j in secondary_graph['node_indices']
                               if (secondary_graph['node_adjacency'][secondary_index, j] or
                                   secondary_graph['node_adjacency'][j, secondary_index])
                               and j in secondary_visited_indices]
    if len(primary_visited_edges) != len(secondary_visited_edges):
        raise StopIteration('Not the same number of edges into the set of already visited nodes')

    #print(f'ENTER === primary: {primary_index} --- secondary: {secondary_index}')
    primary_node_count = primary_graph['node_indices'].shape[0]
    primary_node_mask = np.zeros(shape=(primary_node_count, ), dtype=bool)
    primary_adjacency_mask = np.zeros(shape=(primary_node_count, primary_node_count), dtype=bool)

    for j in secondary_graph['node_indices']:

        secondary_outgoing = secondary_graph['node_adjacency'][secondary_index, j]
        secondary_ingoing = secondary_graph['node_adjacency'][j, secondary_index]
        #print(f'j: {j}   primary ignore {primary_ignore_indices} secondary ignore: {secondary_ignore_indices} ({secondary_ingoing}, {secondary_outgoing})')
        if any([j in secondary_visited_indices,
                not secondary_ingoing and not secondary_outgoing]):
            continue

        secondary_node_attributes = secondary_graph['node_attributes'][j]

        # By default we will assume that such a node does not exist in the local neighborhood of the primary
        # graph and have to be proven wrong first by inspecting each one of them
        exists = False
        for i in primary_graph['node_indices']:

            primary_outgoing = bool(primary_graph['node_adjacency'][primary_index, i])
            primary_ingoing = bool(primary_graph['node_adjacency'][i, primary_index])
            if any([i == primary_index,
                    i in primary_visited_indices,
                    not (primary_ingoing and secondary_ingoing),
                    not (primary_outgoing and secondary_outgoing)]):
                continue

            primary_node_attributes = primary_graph['node_attributes'][i]

            nodes_similar = similarity_func(primary_node_attributes, secondary_node_attributes)
            if nodes_similar:
                #print(f'j: {j} - i: {i} (similar: {nodes_similar})  primary ignore: {primary_ignore_indices}')
                try:
                    _node_mask, _adjacency_mask = graphs_locally_similar(
                        primary_graph=primary_graph,
                        primary_index=i,
                        secondary_graph=secondary_graph,
                        secondary_index=j,
                        similarity_func=similarity_func,
                        primary_visited_indices=primary_visited_indices + [primary_index],
                        secondary_visited_indices=secondary_visited_indices + [secondary_index]
                    )
                    exists = True

                    primary_node_mask = np.logical_or(primary_node_mask, _node_mask)
                    primary_node_mask[primary_index] = True
                    primary_node_mask[i] = True

                    primary_adjacency_mask = np.logical_or(primary_adjacency_mask, _adjacency_mask)
                    primary_adjacency_mask[primary_index][i] = primary_outgoing
                    primary_adjacency_mask[i][primary_index] = primary_ingoing
                    break

                except StopIteration as e:
                    pass

        # If we have not found a similar node at the end of the iteration, then there is none within the
        # primary graph and we can declare the two graphs dissimilar altogether
        if not exists:
            raise StopIteration(f'Graphs not similar! No similarity for secondary node #{j} found within '
                                f'local neighborhood of primary graph node #{primary_index}.')

    #print(f'EXIT  === primary: {primary_index} --- secondary: {secondary_index}')
    return primary_node_mask, primary_adjacency_mask


def contains_subgraph(subgraph: dict,
                      graph: dict,
                      similarity_func: Callable = lambda a1, a2: np.all(a1 == a2),
                      subgraph_index: int = 0
                      ) -> Tuple[bool, Optional[np.ndarray], Optional[np.ndarray]]:
    # First, we take the first node of the subgraph and iterate through the entire main graph to see
    # if there even are nodes which match those attributes
    possible_seed_indices = []
    for node_index in graph['node_indices']:
        similar = similarity_func(graph['node_attributes'][node_index],
                                  subgraph['node_attributes'][subgraph_index])
        if similar:
            possible_seed_indices.append(node_index)

    # The list "possible_seed_indices" now contains all the node indices of the main graph, which are
    # possibly also part of the given subgraph structure. Now, for every one of those nodes we explore the
    # local neighborhood of the main graph and the subgraph in parallel to see if they are indeed similar.
    similar = False
    node_mask = None
    adjacency_mask = None
    for node_index in possible_seed_indices:
        try:
            node_mask, adjacency_mask = graphs_locally_similar(
                primary_graph=graph,
                primary_index=node_index,
                secondary_graph=subgraph,
                secondary_index=subgraph_index,
                similarity_func=similarity_func
            )
            similar = True
        except StopIteration:
            pass

    return similar, node_mask, adjacency_mask


def create_colors_eye_tracking_dataset(graph_infos: Dict[str, dict],
                                       dest_path: str,
                                       image_width: int = 900,
                                       image_height: int = 900,
                                       layout_cb: Callable[[nx.Graph], List[int]] =
                                            nx.layout.kamada_kawai_layout,
                                       draw_node_cb: Callable[[plt.Axes, tc.GraphDict, int], None] =
                                            draw_node_color_scatter,
                                       draw_edge_cb: Callable[[plt.Axes, tc.GraphDict, int, int], None] =
                                            draw_edge_black,
                                       dpi: int = 100,
                                       image_extension: str = 'png'
                                       ):
    # First of all we need to make sure that the given dest_path is actually a directory as it should be
    assert os.path.exists(dest_path), (f'The destination path for the eye tracking dataset {dest_path} does '
                                       f'not exist! Make sure to create the folder!')
    assert os.path.isdir(dest_path), (f'The destination path for the eye tracking dataset "{dest_path}" '
                                      f'is not a directory, but it has to be!')

    # For every element in the given dict we create one eye tracking dataset element, which consists of two
    # files: An image which is a visual representation of the graph itself which will be shown to the user
    # and a metadata json file for each image.
    for key, info in graph_infos.items():
        g = info['graph']
        name = str(key)

        # First of all we will generate the eye tracking image file which provides a visual representation
        # of the graph
        image_name = f'{name}.{image_extension}'
        image_path = os.path.join(dest_path, image_name)
        #print(image_path)
        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(image_width / dpi, image_height / dpi))
        draw_extended_colors_graph(
            ax=ax,
            g=g,
            layout_cb=layout_cb,
            draw_node_cb=draw_node_cb,
            draw_edge_cb=draw_edge_cb,
        )
        ax.axis('off')
        fig.savefig(fname=image_path, dpi=dpi)
        g['node_coordinates'] = np.array([ax.transData.transform(pos) for pos in g['node_positions']])
        plt.close(fig)

        # Then we create the metadata json file
        metadata_name = f'{name}.json'
        metadata_path = os.path.join(dest_path, metadata_name)
        metadata = {
            **info,
            'image_width': image_width,
            'image_height': image_height,
            'image_path': f'./{image_name}'
        }
        # At this point the 'graph' which is contained in the passed information is still in the normal
        # representation where all the values are numpy arrays. This will cause problems for the json
        # serialization, which is why we use this helper method which converts all the values into nested
        # list representations of those arrays.
        metadata['graph'] = graph_dict_to_list_values(metadata['graph'])
        with open(metadata_path, mode='wb') as json_file:
            content = json.dumps(metadata)
            json_file.write(content)




def generate_colors_graph(node_count: int,
                          additional_edge_count: int,
                          colors: List[Tuple[float, float, float]],
                          color_weights: Optional[List[float]] = None,) -> Dict[str, np.ndarray]:
    # If no explicit weights for the random color choice are provided, we assume that all colors should
    # occur with equal probability.
    if color_weights is None:
        color_weights = [1] * len(colors)

    node_indices = list(range(node_count))
    node_attributes = [] + random.choices(colors, weights=color_weights, k=1)
    edge_indices = []

    node_indices_inserted = [0]
    node_indices_remaining = list(range(1, node_count))
    random.shuffle(node_indices_remaining)

    # ~ Generating basic graph structure
    while len(node_indices_remaining) != 0:
        node_1 = random.choice(node_indices_inserted)
        node_2 = node_indices_remaining.pop()

        # Generating the color attributes for the newly inserted node
        node_attributes += random.choices(colors, weights=color_weights, k=1)

        # Adding an undirected edge between the two nodes to the edge list (= two directed edges)
        edge_indices += [(node_1, node_2), (node_2, node_1)]

        # In the end we also have to actually add the node to the list of inserted nodes
        node_indices_inserted.append(node_2)

    # ~ Adding the additional edges
    inserted_edge_count = 0
    while inserted_edge_count != additional_edge_count:
        node_1, node_2 = random.sample(node_indices, 2)

        # If no edge between these two nodes already exists, we add one
        if (node_1, node_2) not in edge_indices and (node_2, node_1) not in edge_indices:
            edge_indices += [(node_1, node_2), (node_2, node_1)]
            inserted_edge_count += 1

    # ~ Creating edge weights and adjancency matrix
    edge_attributes = np.ones(shape=(len(edge_indices), 1))
    node_adjacency = [[1 if (i, j) in edge_indices else 0 for i in node_indices] for j in node_indices]

    # ~ Converting to numpy arrays
    return {
        'node_indices':             np.array(node_indices, dtype=np.int32),
        'node_attributes':          np.array(node_attributes, dtype=np.float32),
        'node_adjacency':           np.array(node_adjacency, dtype=np.int32),
        'edge_indices':             np.array(edge_indices, dtype=np.int32),
        'edge_attributes':          np.array(edge_attributes, dtype=np.float32)
    }


class ExtendedColorsGraphGenerator:
    """
    This class can be used to randomly generate graphs of the so called "EXTENDED COLORS" type.

    Extended Colors
    ---------------
    This type of graphs is mainly identified by the fact that each node of the graph is assigned a color as
    it's main attribute. The first three values of the node attribute vectors will always define the [0, 1]
    RGB values which define the color of that node. It is additionally called "extended" because there is
    also the possibility to generate additional attributes for the graphs besides those colors. This could
    for example be an additional binary attribute which determines if the node is a circle or a square.
    All in all, this kind of dataset is focused on rich & intuitive visual representations of graphs, as it's
    purpose is to be presented to human participants in a visual classification task.

    Generation Procedure
    --------------------
    The generation procedure is a seeding process. The most simple version works like this: A fixed number
    of nodes has to be provided. All of them are initialized as "not inserted", a random node is determined
    as seed node and inserted. From that point onwards a random node of the "not inserted" is chosen and
    inserted as neighbor to one randomly chosen already "inserted" node. This is repeated until all nodes
    are inserted and it results in a "tree" graph. To make cycles possible an additional given number of
    additional edges is inserted between random pairs of nodes afterwards as well. Node properties like the
    color and edge properties as well are chosen according to the callback functions which are passed to the
    generation procedure for that purpose.

    There is also the possibility to use entire graphs as the seeds for the generation process. Multiple
    graphs can be passed to the constructor in the format of a list of GraphDict's. If multiple seeding
    subgraph structures are provided, multiple non connected trees will be grown from those at first and
    then in the end connected by the additional edges.

    Example
    -------

    .. code-block:: python

        import random
        from gnn_student_teacher.data import ExtendedColorsGraphGenerator

        generator = ExtendedColorsGraphGenerator(
            node_count=15,
            additional_edge_count=2,
            color_attributes_cb=lambda: random.choice([[1, 0, 0], [0, 1, 0], [0, 0, 1]]),
            additional_attributes_cb=lambda: [],
            edge_attributes_cb=lambda a, b: [1]
        )

    """
    def __init__(self,
                 node_count: int,
                 additional_edge_count: int,
                 color_attributes_cb: Callable[[tc.GraphDict, int], list],
                 additional_attributes_cb: Callable[[tc.GraphDict, int], List[float]],
                 edge_attributes_cb: Callable[[tc.GraphDict, int, int], List[float]],
                 edge_valid_cb: Callable[[tc.GraphDict, int, int], bool] = lambda g, i, j: True,
                 seed_graphs: List[tc.GraphDict] = [],
                 is_directed: bool = False,
                 prevent_edges_in_seed_graphs: bool = True):
        assert additional_edge_count >= len(seed_graphs) - 1, \
            (f'The "additional_edge_count" ({additional_edge_count}) has to be at least as high as the '
             f'number of provided seed graphs - 1 ({len(seed_graphs) - 1}) to ensure connectivity!')

        self.node_count = node_count
        self.additional_edge_count = additional_edge_count
        self.color_attributes_cb = color_attributes_cb
        self.additional_attributes_cb = additional_attributes_cb
        self.edge_attributes_cb = edge_attributes_cb
        self.edge_valid_cb = edge_valid_cb
        self.seed_graphs = seed_graphs
        self.is_directed = is_directed
        self.prevent_edges_in_seed_graphs = prevent_edges_in_seed_graphs

        # ~ Computed properties
        # These properties about the graph can be derived from the information we are given

        # This is the number of edges which are contributed by the seed graphs. If there are no seed
        # graphs then we set that to -1 because then in that case the graph is grown from a single seed
        # node, which in itself does not contribute an edge to the graph which is actually removing one
        # edge under the assumption we make that in the graph growing process every node contributes one
        # edge to the node
        self.seed_node_count = sum(len(g['node_indices']) for g in seed_graphs)
        if seed_graphs:
            self.seed_edge_count = sum(len(g['edge_indices']) for g in seed_graphs)
        else:
            if self.is_directed:
                self.seed_edge_count = -1
            else:
                self.seed_edge_count = -2
        # undirected graphs will be assumed to just be directed graphs where nodes are connected by two
        # edges pointing both ways.
        if self.is_directed:
            self.edge_count = self.seed_edge_count + \
                              (self.node_count - self.seed_node_count) + self.additional_edge_count
        else:
            self.edge_count = self.seed_edge_count + \
                              ((self.node_count - self.seed_node_count) + self.additional_edge_count) * 2

        # ~ Graph properties
        # these following values are the array-like data structures which have to be constructed and which
        # directly represent the graph that is to be generated
        self.node_indices: List[int] = list(range(node_count))
        self.node_attributes: List[List[float]] = []
        self.node_adjacency: List[List[int]] = []
        self.edge_indices: List[List[int]] = []
        self.edge_attributes: List[List[float]] = []

        self.graph: Dict[str, list] = {}

        # ~ Generation properties
        # These properties are helpers which are needed for the construction process

        self.node_indices_inserted: List[int] = []
        self.node_indices_remaining: List[int] = []
        # The keys of this dict will be the indices of the seed graphs in the given list, the values will
        # be mappings which map the index internal to the seed graph to an index of this graph.
        self.seed_graph_index_maps: Dict[int, Dict[int, int]] = {}
        # This is a list where the index is the node index and the value is the index of the seed graph
        # from which that node was grown.
        self.seed_graph_association: List[int] = []
        self.seed_graph_indices: List[Optional[int]] = []
        # To keep track of the index where to insert the next edge element into edge_indices and
        # edge_attributes
        self.current_edge_index = 0

    def reset(self):
        # node_indices never get modified thats why they dont need to get reset here
        # We initialize all the property lists here with essentially empty items because during the
        # construction process we want to use indiced assignment at certain positions and that is not
        # possible with empty lists
        self.node_attributes = [[] for _ in range(self.node_count)]
        self.node_adjacency = [[0 for _ in range(self.node_count)] for _ in self.node_indices]
        self.edge_indices = [[] for _ in range(self.edge_count)]
        self.edge_attributes = [[] for _ in range(self.edge_count)]

        # Resetting the computation state
        self.node_indices_remaining = copy(self.node_indices)
        # We shuffle the node index list here already so that we get random node choices by simply popping
        # the first element in the other steps of the construction process.
        random.shuffle(self.node_indices_remaining)
        self.node_indices_inserted = []
        self.seed_graph_association = [0 for _ in range(self.node_count)]
        self.seed_graph_indices = [-1 for _ in range(self.node_count)]
        self.current_edge_index = 0

        self.graph = {
            'node_indices': self.node_indices,
            'node_attributes': self.node_attributes,
            'edge_indices': self.edge_indices,
            'edge_attributes': self.edge_attributes,
            'seed_graph_indices': self.seed_graph_indices
        }

    def generate(self) -> tc.GraphDict:
        self.reset()

        # ~ Seeding
        # The whole generation procedure is a seeding method. The graph starts out with a seed center and
        # the rest of the graph is then "grown" around that. It is possible to define subgraphs which act
        # as seeds. If that is not the case, then one node is chosen at random to be the seed
        if len(self.seed_graphs) == 0:
            self.insert_seed_node()
        else:
            self.insert_seed_graphs()

        # ~ Growing
        self.grow_nodes()
        self.add_edges()

        self.node_adjacency = self.node_adjacency_from_edge_indices(self.node_count, self.edge_indices)

        return {
            'node_indices':         np.array(self.node_indices, dtype=np.int32),
            'node_attributes':      np.array(self.node_attributes, dtype=np.float32),
            'node_adjacency':       np.array(self.node_adjacency, dtype=np.int32),
            'edge_indices':         np.array(self.edge_indices, dtype=np.int32),
            'edge_attributes':      np.array(self.edge_attributes, dtype=np.float32),
            'seed_graph_indices':   np.array(self.seed_graph_indices, dtype=np.int32),
        }

    def insert_seed_graphs(self):
        for graph_index, g in enumerate(self.seed_graphs):
            seed_size = len(g['node_indices'])
            seed_indices = random.sample(self.node_indices_remaining, k=seed_size)
            # This is a mapping whose keys are the node indices local to the seed graph and the values are
            # the corresponding values are the node indices chosen for this graph to be generated
            seed_index_map = dict(zip(g['node_indices'], seed_indices))
            self.seed_graph_index_maps[graph_index] = seed_index_map

            for seed_index, index in seed_index_map.items():
                self.node_attributes[index] = g['node_attributes'][seed_index]
                self.seed_graph_association[index] = graph_index
                self.seed_graph_indices[index] = graph_index

                # Technically the node indices are now inserted, now we need to establish the corresponding
                # edge connections according to the seed graph.
                for (i, j), edge_attributes in zip(g['edge_indices'], g['edge_attributes']):
                    if seed_index == i:
                        edge = [seed_index_map[i], seed_index_map[j]]
                        self.edge_indices[self.current_edge_index] = edge
                        self.edge_attributes[self.current_edge_index] = edge_attributes
                        self.current_edge_index += 1

                # We now have to remove this node from the list of remaining nodes and also have to add it
                # to the list of inserted nodes
                self.node_indices_remaining.remove(index)
                self.node_indices_inserted.append(index)

    def create_single_node_attributes(self, index: int) -> List[float]:
        color_attributes = self.color_attributes_cb(self.graph, index)
        additional_attributes = self.additional_attributes_cb(self.graph, index)
        return color_attributes + additional_attributes

    def create_single_edge_attributes(self, i: int, j: int) -> List[float]:
        return self.edge_attributes_cb(self.graph, i, j)

    def insert_seed_node(self):
        index = self.node_indices_remaining.pop(0)
        self.node_attributes[index] = self.create_single_node_attributes(0)
        self.node_indices_inserted.append(index)

    def insert_edge(self, i: int, j: int) -> None:
        if self.is_directed:
            edges = [[i, j]]
        else:
            edges = [[i, j], [j, i]]

        for k, l in edges:
            self.edge_indices[self.current_edge_index] = [k, l]
            self.edge_attributes[self.current_edge_index] = self.create_single_edge_attributes(k, l)
            self.current_edge_index += 1

    def grow_nodes(self):
        while len(self.node_indices_remaining) > 0:
            index = self.node_indices_remaining.pop(0)

            # now we choose a random already inserted node to attach this node to
            anchor_index = random.choice(self.node_indices_inserted)

            self.node_attributes[index] = self.create_single_node_attributes(anchor_index)
            self.seed_graph_association[index] = self.seed_graph_association[anchor_index]
            self.insert_edge(index, anchor_index)
            self.node_indices_inserted.append(index)

    def add_edges(self):
        # This method is supposed to add additional edges to the grown graph, because the base growing
        # procedure will always be a tree (no cycles).
        # The most important part of this will be to connect the multiple non-connected graphs grown from
        # potentially multiple seed graphs. All additional edges beyond that will be inserted between
        # random nodes.
        additional_edge_count = self.additional_edge_count
        if len(self.seed_graphs) > 1:
            for graph_index in range(len(self.seed_graphs) - 1):
                if (len(self.seed_graphs[graph_index]['node_indices']) == 0 or
                        len(self.seed_graphs[graph_index + 1]['node_indices']) == 0):
                    continue
                # We choose a random node from both subgraph seeded clusters of nodes and insert an edge
                # there
                i = random.choice([k for k, gi in enumerate(self.seed_graph_association)
                                   if gi == graph_index and self.seed_graph_indices[k] < 0])
                j = random.choice([k for k, gi in enumerate(self.seed_graph_association)
                                   if gi == graph_index + 1])
                self.insert_edge(i, j)
                additional_edge_count -= 1

        while additional_edge_count > 0:
            i, j = random.sample(self.node_indices_inserted, k=2)

            # This first condition here makes sure that the edge is only inserted when that exact same
            # edge does not already exist.
            edge_index_tuples = [tuple(e) for e in self.edge_indices]
            edge_exist = ((i, j) in edge_index_tuples or
                          (j, i) in edge_index_tuples)
            # The second condition makes sure that if the prevent_edges_in_seed_graphs flag is set that we
            # do not insert an additional edge between two nodes of the seed graph, as that would kind of
            # disturb the subgraph structure
            seed_cond = (self.seed_graph_indices[i] == -1 or
                         self.seed_graph_indices[j] == -1 or
                         not self.prevent_edges_in_seed_graphs)

            # It is additionally possible to pass a callback to the constructor of this class which takes a
            # possible edge as arguments and then depending on some criteria returns the boolean value of
            # whether this edge is to be considered a valid insertion or not
            valid_cond = self.edge_valid_cb(self.graph, i, j)

            if not edge_exist and seed_cond and valid_cond:
                self.insert_edge(i, j)
                additional_edge_count -= 1

    @classmethod
    def node_adjacency_from_edge_indices(cls,
                                         node_count: int,
                                         edge_indices: List[List[int]]) -> List[List[int]]:
        node_adjacency = [[0 for _ in range(node_count)] for _ in range(node_count)]
        for i, j in edge_indices:
            node_adjacency[i][j] = 1

        return node_adjacency



"""
DEPRECATED - The contents of this module have been split into individual modules within the "models"
package.
"""
from typing import List, Optional, Callable, Tuple, Union
import time
import logging
import typing as t

import numpy as np
import tensorflow as tf
import tensorflow.keras as ks
from tensorflow.python.keras.engine import compile_utils
from kgcnn.data.utils import ragged_tensor_from_nested_numpy
from kgcnn.layers.modules import GraphBaseLayer
from kgcnn.layers.modules import LazyConcatenate, LazyAverage
from kgcnn.layers.modules import DenseEmbedding, ActivationEmbedding, DropoutEmbedding
from kgcnn.layers.pooling import PoolingLocalEdges
from kgcnn.layers.pooling import PoolingNodes
from kgcnn.layers.pooling import PoolingWeightedNodes
from kgcnn.layers.conv.gcn_conv import GCN
from kgcnn.layers.conv.gat_conv import AttentionHeadGAT, AttentionHeadGATV2
from kgcnn.literature.GNNExplain import GNNInterface

from graph_attention_student.layers import MultiHeadGatLayer
from graph_attention_student.layers import MultiHeadGATV2Layer
from graph_attention_student.layers import ExplanationSparsityRegularization
from graph_attention_student.training import RecompilableMixin
from graph_attention_student.training import shifted_sigmoid
from graph_attention_student.training import mae, mse
from graph_attention_student.util import NULL_LOGGER


class MultiAttentionStudent(RecompilableMixin, ks.models.Model):

    def __init__(self,
                 # primary network related arguments
                 units: List[int],
                 activation: str = 'kgcnn>leaky_relu',
                 use_bias: bool = True,
                 dropout_rate: float = 0.0,
                 use_edge_features: bool = True,
                 # importance related arguments
                 importance_units: List[int] = [],
                 importance_channels: int = 2,
                 importance_activation: str = 'sigmoid',
                 importance_dropout_rate: float = 0.0,
                 importance_multiplier: float = 1,
                 importance_sparsity: float = 1e-3,
                 importance_exclusivity: float = 0.0,
                 final_units: List[int] = [],
                 final_dropout_rate: float = 0.0,
                 final_activation: str = 'softmax',
                 importance_factor: float = 0.0,
                 regression_limits: Optional[Union[Tuple[float, float], List[tuple]]] = None,
                 regression_bins: Optional[List[List[float]]] = None,
                 regression_reference: float = 0,
                 sparsity_factor: float = 0.0,
                 **kwargs):
        ks.models.Model.__init__(self, **kwargs)
        RecompilableMixin.__init__(self)

        self.units = units
        self.activation = activation
        self.use_bias = use_bias
        self.dropout_rate = dropout_rate
        self.use_edge_features = use_edge_features

        self.regression_reference = regression_reference
        self.regression_limits: Optional[Tuple[float, float]] = regression_limits
        self.regression_bins = regression_bins
        self.regression_center: Optional[float] = None

        self.importance_multiplier = importance_multiplier
        self.importance_channels = importance_channels
        self.importance_activation = importance_activation
        self.importance_dropout_rate = importance_dropout_rate
        self.final_units = final_units
        self.final_activation = final_activation
        self.importance_factor = importance_factor
        self.importance_exclusivity = importance_exclusivity
        if isinstance(importance_sparsity, float):
            self.importance_sparsity = [importance_sparsity for _ in range(importance_channels)]
        elif isinstance(importance_sparsity, list):
            self.importance_sparsity = importance_sparsity

        # ~ VALIDATING ARGUMENTS ~

        # ~ NETWORK COMPONENTS ~

        # This list contains all the actual GAT-like layers which do the graph convolutions. How many
        # layers there will be and how many hidden units they have is determined by the "self.units" list.
        self.attention_layers: List[GraphBaseLayer] = []
        for units in self.units:
            # This is a custom layer, which is essentially a multi-head GATv2, just compiled into a single
            # layer.
            lay = MultiHeadGatLayer(
                units=units,
                num_heads=importance_channels,
                use_edge_features=use_edge_features,
                activation=activation,
                use_bias=use_bias,
                use_final_activation=True,
                has_self_loops=True,
                share_weights=False,
                # If this is True the embeddings of the heads are concat-ed and then passed to the input of
                # the next layer. If it is False, they are averaged instead
                concat_heads=True,
            )
            self.attention_layers.append(lay)

        # various layers just for regularization
        self.lay_dropout = DropoutEmbedding(rate=dropout_rate)
        self.lay_dropout_importances = DropoutEmbedding(rate=importance_dropout_rate)

        self.lay_act_importance = ActivationEmbedding(importance_activation)

        # ~ Edge Importances
        # The edge importances are derived from the edge attention weights which the GAT layers maintain
        # anyways. Since we have as many parallel GAT "heads" as there are to be importance channels, we
        # only have to reduce over the layer-dimension to get the edge importance values.
        self.lay_concat_alphas = LazyConcatenate(axis=-1)

        self.lay_pool_edges_in = PoolingLocalEdges(pooling_method='mean', pooling_index=0)
        self.lay_pool_edges_out = PoolingLocalEdges(pooling_method='mean', pooling_index=1)
        self.lay_average = LazyAverage()

        # ~ Node Importances
        # The node importances are calculated from an additional series of dense layer which works on
        # the final node embedding that was produced by all of the previously defined GAT layers together.

        # The final unit count has to be exactly the number of importance channels.
        self.node_importance_units = importance_units + [importance_channels]
        # The final activation is a relu here, which might seem weird because we additionally also use an
        # additional sigmoid activation on this already relu-ed value. This is actually really important
        # and further explained in the docstring.
        self.node_importance_acts = ['relu' for _ in importance_units] + ['linear']
        self.node_importance_layers = []
        for k, act in zip(self.node_importance_units, self.node_importance_acts):
            lay = DenseEmbedding(
                units=k,
                activation=act,
                use_bias=use_bias,
            )
            self.node_importance_layers.append(lay)

        # ~ Output
        # The main GAT layers produce final node embeddings for the graph at the end
        # These node embeddings are then globally pooled into a vector of graph embeddings - actually one
        # such vector of graph embeddings for each importance channel. Those are then concat-ed and fed into
        # a final series of Dense layers, which then produce the final output...
        self.lay_pool_out_weighted = PoolingWeightedNodes(pooling_method='sum')
        self.lay_pool_out = PoolingNodes(pooling_method='sum')

        self.lay_concat_out = LazyConcatenate(axis=-1)
        self.lay_final_dropout = DropoutEmbedding(final_dropout_rate)

        if self.doing_regression:
            self.out_units = final_units + [1]
            self.out_activations = ['relu' for _ in final_units] + ['linear']
            self.lay_final_activation = ActivationEmbedding('linear')

        elif self.doing_classification:
            self.out_units = final_units + [self.importance_channels]
            self.out_activations = ['relu' for _ in final_units] + ['linear']
            self.lay_final_activation = ActivationEmbedding('softmax')

        self.out_layers = []
        for k, act in zip(self.out_units, self.out_activations):
            lay = DenseEmbedding(
                units=k,
                activation=act,
                use_bias=use_bias,
                # use_bias=False
            )
            self.out_layers.append(lay)

        # tbd
        self.bce_loss = ks.losses.BinaryCrossentropy()
        self.compiled_bce_loss = compile_utils.LossesContainer(self.bce_loss)

        self.cce_loss = ks.losses.CategoricalCrossentropy()
        self.compiled_cce_loss = compile_utils.LossesContainer(self.cce_loss)

        self.mse_loss = ks.losses.MeanSquaredError()
        self.compiled_mse_loss = compile_utils.LossesContainer(self.mse_loss)

        self.lay_exp = DenseEmbedding(
            units=self.importance_channels,
            use_bias=False,
            #kernel_constraint=ks.constraints.NonNeg(),
            kernel_initializer=ks.initializers.Identity(),
        )
        self.exp_layers = []
        for k in range(self.importance_channels):
            lay = DenseEmbedding(
                units=1,
                activation='relu',
                use_bias=True
            )
            self.exp_layers.append(lay)

        if self.doing_classification:
            self.lay_final = DenseEmbedding(units=self.importance_channels, activation='softmax')
        elif self.doing_regression:
            self.lay_final = DenseEmbedding(units=1, activation='linear')

        # By default the network is built to do multi-class graph classification, but it is also possible
        # to do regression, but that requires some special changes.
        if self.regression_limits is not None:
            self.regression_center = self.regression_reference
            self.regression_width = np.abs(self.regression_limits[1] - self.regression_limits[0])

        self.lay_sparsity = ExplanationSparsityRegularization(coef=sparsity_factor)

    # def build(self, *args, **kwargs):
    #     print('BUILD', args, kwargs)
    #     super(MultiAttentionStudent, self).build(*args, **kwargs)

    @property
    def doing_regression(self) -> bool:
        return self.regression_limits is not None

    @property
    def doing_classification(self) -> bool:
        return self.regression_bins is None

    def finalize(self, out):
        pass

    def call(self,
             inputs,
             training=False,
             external_node_importances: Optional[np.ndarray] = None,
             mask_channel: Optional[int] = None,
             **kwargs):
        # node_input: ([batch], [N], V)
        # edge_input: ([batch], [M], F)
        # edge_index_input: ([batch], [M], 2)
        node_input, edge_input, edge_index_input = inputs
        #print(node_input.shape, edge_input.shape, edge_index_input.shape)

        # Then we pass the input through all those multi channel gat layers. Basically each channel maintains
        # it's own node feature embeddings. Those are only combined into a single node embedding vector at
        # the very end. Throughout all the layers we also collect the edge attention coefficient "alpha"
        # vector
        alphas = []
        xs = []
        x = node_input
        for lay in self.attention_layers:
            # x: ([batch], [N], V)
            # alpha: ([batch], [M], K, 1)
            x, alpha = lay([x, edge_input, edge_index_input])
            if training:
                x = self.lay_dropout(x, training=training)

            xs.append(x)
            alphas.append(alpha)

        # ~ edge importances
        # alphas: ([batch], [M], K, L) - where L is number of layers. This is also the dimension we reduce
        alphas = self.lay_concat_alphas(alphas)
        alphas = tf.reduce_sum(alphas, axis=-1, keepdims=False)
        # edge_importances: ([batch], [M], K)
        edge_importances = alphas
        # We want importances to also be a [0, 1] attention-like value
        edge_importances = ks.activations.sigmoid(edge_importances)

        # Now we need to pool the edge importances, which essentially turns them into node embeddings.

        # The first argument "node_input" we pass here is completely arbitrary. It does not even have
        # an effect.
        pooled_edges_in = self.lay_pool_edges_in([node_input, edge_importances, edge_index_input])
        pooled_edges_out = self.lay_pool_edges_out([node_input, edge_importances, edge_index_input])
        # pooled_edges: ([batch], [N], K)
        pooled_edges = self.lay_average([pooled_edges_in, pooled_edges_out])

        # ~ node_importances
        node_importances_raw = tf.concat([x], axis=-1)
        for lay in self.node_importance_layers:
            node_importances_raw = lay(node_importances_raw)

        node_importances = ks.activations.sigmoid(node_importances_raw)
        node_importances_slices = []

        # This multiplication of node_importances and pooled edge_importances helps to make the explanation
        # more "consistent" -> There will likely be not too many freely floating node / edge importances.
        # Like this it is more likely that node and edges which are directly besides each other have similar
        # importance values.
        node_importances = node_importances * pooled_edges

        for k in range(self.importance_channels):
            importances_slice = tf.expand_dims(node_importances[:, :, k], axis=-1)
            node_importances_slices.append(importances_slice)

        if mask_channel is not None:
            node_importances_slices = [tf.zeros_like(s) if k != mask_channel else s
                                       for k, s in enumerate(node_importances_slices)]

        node_importances = tf.concat(node_importances_slices, axis=-1)
        self.lay_sparsity(node_importances)
        node_importances = self.lay_dropout_importances(node_importances)

        if external_node_importances is not None:
            node_importances = external_node_importances

        # ~ output
        outs = []
        for k in range(self.importance_channels):
            # For each channel we use the corresponding importance vector and use it as the weights to
            # perform a weighted global sum pooling on the corresponding node embeddings to produce the
            # overall graph embeddings
            node_importance_slice = tf.expand_dims(node_importances[:, :, k], axis=-1)

            # This pooling operation produces global graph embeddings
            # out: ([batch], V, 1)
            out = self.lay_pool_out(x * node_importance_slice)
            # for lay in self.out_layers[k]:
            #     out = lay(out)

            outs.append(out)

        # out: ([batch], K)
        out = self.lay_concat_out(outs)

        for lay in self.out_layers:
            out = lay(out)
            if training:
                out = self.lay_final_dropout(out, training=training)

        logits = out
        out = self.lay_final_activation(logits)

        if self.doing_regression:
            out = tf.squeeze(out, axis=-1)
            out = out + self.regression_reference

        if training:
            return out, logits, x, node_importances, edge_importances
        else:
            return out, node_importances, edge_importances

    def predict_single(self,
                       inputs: Tuple[List[float], List[float], List[float]],
                       external_node_importances: Optional[List[List[float]]] = None):
        node_input, edge_input, edge_index_input = inputs

        if external_node_importances is not None:
            external_node_importances = ragged_tensor_from_nested_numpy(np.array([external_node_importances]))
            external_node_importances = tf.cast(external_node_importances, tf.float32)

        # The input is just normal lists, to be processable by the model we need to turn those into
        # ragged tensors first
        label, node_importances, edge_importances = self([
                ragged_tensor_from_nested_numpy(np.array([node_input])),
                ragged_tensor_from_nested_numpy(np.array([edge_input])),
                ragged_tensor_from_nested_numpy(np.array([edge_index_input]))
            ],
            training=False,
            finalize=True,
            external_node_importances=external_node_importances
        )

        # The results come out as tensors as well, we convert them back to lists.
        return (
            label.numpy().tolist()[0],
            node_importances.to_list()[0],
            edge_importances.to_list()[0]
        )

    def regression_augmentation(self,
                                out_true: tf.RaggedTensor):
        center_distances = tf.abs(out_true - self.regression_center)
        center_distances = (center_distances * self.importance_multiplier) / (0.5 * self.regression_width)

        lo_samples = tf.where(out_true < self.regression_center, center_distances,  0.0)
        lo_samples = tf.expand_dims(lo_samples, axis=-1)
        lo_mask = tf.where(out_true < self.regression_center, 1.0, 0.0)
        lo_mask = tf.expand_dims(lo_mask, axis=-1)

        hi_samples = tf.where(out_true > self.regression_center, center_distances,  0.0)
        hi_samples = tf.expand_dims(hi_samples, axis=-1)
        hi_mask = tf.where(out_true > self.regression_center, 1.0, 0.0)
        hi_mask = tf.expand_dims(hi_mask, axis=-1)

        reg_true = tf.concat([lo_samples, hi_samples], axis=-1)
        mask = tf.concat([lo_mask, hi_mask], axis=-1)
        # reg_true = tf.concat([hi_samples, lo_samples], axis=-1)
        # mask = tf.concat([hi_mask, lo_mask], axis=-1)

        return reg_true, mask

    def create_regression_mask(self,
                               out_true: tf.RaggedTensor):

        mask = []
        for k, (lower, upper) in enumerate(self.regression_bins):
            mask_slice = tf.where(out_true <= upper, 1.0, 0.0) * tf.where(out_true >= lower, 1.0, 0.0)
            mask_slice = tf.expand_dims(mask_slice, axis=-1)
            mask.append(mask_slice)

        mask = tf.concat(mask, axis=-1)

        return mask

    def train_step_explanation(self,
                               x: Tuple[tf.RaggedTensor, tf.RaggedTensor, tf.RaggedTensor],
                               y: Tuple[tf.RaggedTensor, tf.RaggedTensor, tf.RaggedTensor],
                               ) -> None:
        """
        Given the input of one batch ``x`` and the corresponding ground truth values ``y`` performs a single
        complete train step for the explanation. Calculates a gradient and applies changes immediately to
        the trainable parameters.

        Generally, the explanation is trained by implicitly assuming that the GAT part of the
        network outputs a constant value of 1 for every node embedding. With this assumption the same global
        pooling is performed to create graph embeddings for every explanation channel through a weighted
        global pooling with the corresponding node importance channel values. The additional output dense
        layers are then discarded for this step.

        These graph embeddings are then used to solve an augmented problem, which depends on whether the
        network is doing regression or classification.

        **CLASSIFICATION:**

        The augmented problem uses the same ground truth class labels. The previously described graph
        embeddings for each channel are submitted to a modified sigmoid activation and then a BCE loss is
        used to train each channel to individually predict one of the classes.

        **REGRESSION:**

        For regression we first need a-priori knowledge about the range of values to be expected from the
        dataset. From that we can deduce the center of that value range. By default we map a regression
        problem to be composed of 2 channels: One which explains high values and one which explains low
        values (relative to the center of the expected value range).
        We then set up two separate regression problems by only considering the absolute distance from this
        center value. One of the channels is then only trained on all the samples which have values below
        the center value and the other is only trained with the samples which have original values above
        this center value.

        .. code-block:: text

            # pseudocode
            lo_true = abs(y_true - center) where y_true < center
            hi_true = abs(y_true - center) where y_true >= center

            reg_true = concat(lo_true, hi_true)

        The predictions of the network are computed by applying a relu activation on the previously described
        simplified graph embeddings of each channel and then using MSE loss to train on the vector of
        constructed regression values ``reg_true``.

        :param x: Tuple consisting of the 3 required input tensors: <br>
            - node_input: ragged tensor of node features ([batch], [V], N0) <br>
            - edge_input: ragged tensor of edge features ([batch], [E], M0) <br>
            - edge_indices: ragged tensor of edge index tuples ([batch], [E], 2)
        :param y: Tuple consisting of ground truth target tensors: <br>
            - out_true: ragged tensor of shape ([batch], ?). Final dimension depends on whether the network
              is used for regression or classification. <br>
            - ni_true: ragged tensor of node importances ([batch], [V], K) <br>
            - ei_true: ragged tensor of edge importances ([batch], [E], K)
        :return: None
        """
        node_input, edge_input, edge_index_input = x
        out_true, ni_true, ei_true = y

        with tf.GradientTape() as tape:
            y_pred = self(x, training=True)  # Forward pass
            out_pred, logits_pred, x_pred, ni_pred, ei_pred = y_pred

            outs = []
            for k in range(self.importance_channels):
                node_importance_slice = tf.expand_dims(ni_pred[:, :, k], axis=-1)

                out = self.lay_pool_out(node_importance_slice)
                # out = self.lay_pool_out(node_importance_slice * x_pred)
                # out = tf.reduce_sum(out, keepdims=1)

                #out = self.exp_layers[k](out)

                outs.append(out)

            out = self.lay_concat_out(outs)

            if self.doing_regression:

                reg_true, mask = self.regression_augmentation(out_true)
                reg_pred = out

                exp_loss = (self.importance_channels) * \
                           self.compiled_mse_loss(reg_true * mask, reg_pred * mask)

                exp_loss += (self.importance_sparsity[k] / self.importance_channels) \
                    * tf.reduce_mean(tf.abs(ni_pred[:, :, k]))

            elif self.doing_classification:
                class_pred = shifted_sigmoid(out, multiplier=self.importance_multiplier)
                exp_loss = self.compiled_bce_loss(out_true, class_pred)

            exp_loss *= self.importance_factor

        # Compute gradients
        trainable_vars = self.trainable_variables
        exp_gradients = tape.gradient(exp_loss, trainable_vars)

        # Update weights
        self.optimizer.apply_gradients(zip(exp_gradients, trainable_vars))

        return {'exp_loss': exp_loss}

    def train_step_explanation_alt(self,
                                   x: Tuple[tf.RaggedTensor, tf.RaggedTensor, tf.RaggedTensor],
                                   y: Tuple[tf.RaggedTensor, tf.RaggedTensor, tf.RaggedTensor],
                                   ) -> None:
        node_input, edge_input, edge_index_input = x
        out_true, ni_true, ei_true = y

        if self.doing_regression:
            mask_reg = self.create_regression_mask(out_true)

        exp_loss = 0.0
        with tf.GradientTape(persistent=True) as tape:
            ni_pred_combined = []
            for k in range(self.importance_channels):

                y_pred = self(x, training=True, mask_channel=k)  # Forward pass
                out_pred, logits_pred, x_pred, ni_pred, ei_pred = y_pred

                if self.doing_regression:
                    # reg_pred = self.lay_exp(out)
                    # exp_loss += self.compiled_mse_loss(
                    #     tf.expand_dims(reg_true[:, k], axis=-1),
                    #     tf.expand_dims(out_pred, axis=-1),
                    #     sample_weight=mask[:, k]
                    # )
                    exp_loss += self.compiled_mse_loss(
                        tf.expand_dims(out_true, axis=-1),
                        tf.expand_dims(out_pred, axis=-1),
                        sample_weight=mask_reg[:, k]
                    )

                    exp_loss += (self.importance_sparsity[k] / self.importance_channels) \
                        * tf.reduce_mean(tf.abs(ni_pred[:, :, k]))

                    ni_pred_combined.append(tf.expand_dims(ni_pred[:, :, k], axis=-1))

                elif self.doing_classification:
                    out_true = tf.cast(out_true, tf.float32)
                    mask = tf.where(out_true[:, k] > 0.9, 1.0, 0.0)
                    # mask = tf.expand_dims(mask, axis=-1)
                    # exp_loss += self.compiled_cce_loss(out_true, out_pred, sample_weight=mask)

                    class_pred = tf.expand_dims(ks.activations.sigmoid(logits_pred[:, k]), axis=-1)
                    class_true = tf.ones_like(class_pred)

                    exp_loss += self.compiled_bce_loss(class_true, class_pred, sample_weight=mask)
                    exp_loss += (self.importance_sparsity[k] / self.importance_channels) \
                        * tf.reduce_mean(tf.abs(ni_pred[:, :, k]))

                    ni_pred_combined.append(tf.expand_dims(ni_pred[:, :, k], axis=-1))

            ni_pred_combined = tf.concat(ni_pred_combined, axis=-1)
            entropy = -tf.reduce_sum(ni_pred_combined * tf.math.log(tf.abs(ni_pred_combined) + 1e-8), axis=-1)
            exp_loss += self.importance_exclusivity * tf.reduce_mean(entropy)
            exp_loss *= self.importance_factor

        # Compute gradients
        trainable_vars = self.trainable_variables
        exp_gradients = tape.gradient(exp_loss, trainable_vars)

        # Update weights
        self.optimizer.apply_gradients(zip(exp_gradients, trainable_vars))

        return {'exp_loss': exp_loss}

    def train_step(self,
                   data: Tuple[tuple, tuple]
                   ) -> None:
        """
        Performs one training step with the given ``data`` consisting of the input data ``x`` of one batch
        as well as the corresponding ground truth labels ``y``.

        :param data: Tuple consisting of the input data x of a batch as well as the corresponding ground
            truth labels y. <br>
            x: Tuple consisting of the 3 required input tensors: <br>
            - node_input: ragged tensor of node features ([batch], [V], N0) <br>
            - edge_input: ragged tensor of edge features ([batch], [E], M0) <br>
            - edge_indices: ragged tensor of edge index tuples ([batch], [E], 2)
            y: Tuple consisting of ground truth target tensors: <br>
            - out_true: ragged tensor of shape ([batch], ?). Final dimension depends on whether the network
              is used for regression or classification. <br>
            - ni_true: ragged tensor of node importances ([batch], [V], K) <br>
            - ei_true: ragged tensor of edge importances ([batch], [E], K)
        :return: None
        """
        # Unpack the data. Its structure depends on your model and
        # on what you pass to `fit()`.
        if len(data) == 3:
            x, y, sample_weight = data
        else:
            sample_weight = None
            x, y = data

        # ~ EXPLANATION TRAINING
        # "train_explanation" actually performs an entire training step just for the attention "explanation"
        # part of the network. Look up the details in the function. The gist is that it only uses part of
        # the network to produce a simplified output which is then used to create a single training step
        # for the explanation part of the network.
        exp_metrics = {}
        if self.importance_channels != 1 and self.importance_factor != 0:
            exp_metrics = self.train_step_explanation(x, y)
            # exp_metrics = self.train_step_explanation_alt(x, y)
            pass

        # ~ PREDICTION TRAINING
        # This performs a "normal" training step, as it is in the default implementation of the "train_step"
        out_true, ni_true, ei_true = y
        with tf.GradientTape() as tape:
            y_pred = self(x, training=True)  # Forward pass
            out_pred, logits_pred, x_pred, ni_pred, ei_pred = y_pred

            # Compute the loss value.
            # The loss function is configured in `compile()`.
            loss = self.compiled_loss(
                [out_true, ni_true, ei_true],
                [out_pred, ni_pred, ei_pred],
                sample_weight=sample_weight,
                regularization_losses=self.losses,
            )

        # Compute gradients
        trainable_vars = self.trainable_variables
        gradients = tape.gradient(loss, trainable_vars)

        # Update weights
        self.optimizer.apply_gradients(zip(gradients, trainable_vars))

        # Update the metrics.
        # Metrics are configured in `compile()`.
        self.compiled_metrics.update_state(
            [out_true, ni_true, ei_true],
            [out_pred, ni_pred, ei_pred],
            sample_weight=sample_weight
        )

        # Return a dict mapping metric names to current value.
        # Note that it will include the loss (tracked in self.metrics).
        return {
            **{m.name: m.result() for m in self.metrics},
            **exp_metrics
        }


class GnnxGCN(ks.models.Model, GNNInterface):

    def __init__(self,
                 units: List[int],
                 activation: str = 'kgcnn>leaky_relu',
                 pooling_method: str = 'sum',
                 final_units: List[int] = [],
                 final_activation: str = 'linear',
                 outputs: int = 1,
                 **kwargs):
        ks.models.Model.__init__(self, **kwargs)
        GNNInterface.__init__(self)

        self.conv_layers = []
        for k in units:
            lay = GCN(
                units=k,
                activation=activation,
            )
            self.conv_layers.append(lay)

        self.lay_pooling = PoolingNodes(pooling_method=pooling_method)
        self.final_activations = ['relu' for _ in final_units] + [final_activation]
        self.final_units = [k for k in final_units] + [outputs]
        self.final_layers = []
        for k, act in zip(self.final_units, self.final_activations):
            lay = DenseEmbedding(
                units=k,
                activation=act,
            )
            self.final_layers.append(lay)

    def call(self, inputs):
        node_input, edge_input, edge_index_input = inputs

        x = node_input
        for lay in self.conv_layers:
            x = lay([x, edge_input, edge_index_input])

        out = self.lay_pooling(x)
        for lay in self.final_layers:
            out = lay(out)

        return out

    def predict(self, gnn_input):
        node_input, edge_input, edge_index_input = gnn_input

        return self([node_input, edge_input, edge_index_input], training=False)[0]

    def masked_predict(self, gnn_input, edge_mask, feature_mask, node_mask, **kwargs):
        node_input, edge_input, edge_index_input = gnn_input

        edge_mask = tf.RaggedTensor.from_tensor(tf.cast(tf.expand_dims(edge_mask, axis=0), dtype=tf.float32))
        node_mask = tf.RaggedTensor.from_tensor(tf.cast(tf.expand_dims(node_mask, axis=0), dtype=tf.float32))
        # print(edge_input.shape, edge_mask.shape)
        edge_input_masked = tf.cast(edge_input, dtype=tf.float32) * edge_mask
        node_input_masked = tf.cast(node_input, dtype=tf.float32) * node_mask

        return self([node_input_masked, edge_input_masked, edge_index_input], training=False)[0]

    def get_explanation(self, gnn_input, edge_mask, feature_mask, node_mask):
        return node_mask.numpy().tolist(), edge_mask.numpy().tolist()

    def get_number_of_nodes(self, gnn_input):
        node_input, _, _ = gnn_input
        return node_input[0].shape[0]

    def get_number_of_node_features(self, gnn_input):
        node_input, _, _ = gnn_input
        return node_input.shape[2]

    def get_number_of_edges(self, gnn_input):
        _, edge_input, _ = gnn_input
        return edge_input[0].shape[0]


class Megan(ks.models.Model):
    """
    MEGAN: Multi Explanation Graph Attention Network
    This model currently supports graph regression and graph classification problems. It was mainly designed
    with a focus on explainable AI (XAI). Along the main prediction, this model is able to output multiple
    attention-based explanations for that prediction. More specifically, the model outputs node and edge
    attributional explanations (assigning [0, 1] values to ever node / edge of the input graph) in K
    separate explanation channels, where K can be chosen as an independent model parameter.
    """

    def __init__(self,
                 # convolutional network related arguments
                 units: t.List[int],
                 activation: str = "kgcnn>leaky_relu",
                 use_bias: bool = True,
                 dropout_rate: float = 0.0,
                 use_edge_features: bool = True,
                 # node/edge importance related arguments
                 importance_units: t.List[int] = [],
                 importance_channels: int = 2,
                 importance_activation: str = "sigmoid",  # do not change
                 importance_dropout_rate: float = 0.0,  # do not change
                 importance_factor: float = 0.0,
                 importance_multiplier: float = 10.0,
                 importance_transformations: t.Optional[t.List[ks.layers.Layer]] = None,
                 sparsity_factor: float = 0.0,
                 concat_heads: bool = True,
                 separate_explanation_step: bool = False,
                 # mlp tail end related arguments
                 final_units: t.List[int] = [1],
                 final_dropout_rate: float = 0.0,
                 final_activation: str = 'linear',
                 final_pooling: str = 'sum',
                 regression_limits: t.Optional[t.Tuple[float, float]] = None,
                 regression_bins: t.Optional[t.List[t.Tuple[float, float]]] = None,
                 regression_reference: t.Optional[float] = None,
                 return_importances: bool = True,
                 use_graph_attributes: bool = False,
                 **kwargs):
        """
        Args:
            units: A list of ints where each element configures an additional attention layer. The numeric
                value determines the number of hidden units to be used in the attention heads of that layer
            activation: The activation function to be used within the attention layers of the network
            use_bias: Whether the layers of the network should use bias weights at all
            dropout_rate: The dropout rate to be applied after *each* of the attention layers of the network.
            use_edge_features: Whether edge features should be used. Generally the network supports the
                usage of edge features, but if the input data does not contain edge features, this should be
                set to False.
            importance_units: A list of ints where each element configures another dense layer in the
                subnetwork that produces the node importance tensor from the main node embeddings. The
                numeric value determines the number of hidden units in that layer.
            importance_channels: The int number of explanation channels to be produced by the network. This
                is the value referred to as "K". Note that this will also determine the number of attention
                heads used within the attention subnetwork.
            importance_factor: The weight of the explanation-only train step. If this is set to exactly
                zero then the explanation train step will not be executed at all (less computationally
                expensive)
            importance_multiplier: An additional hyperparameter of the explanation-only train step. This
                is essentially the scaling factor that is applied to the values of the dataset such that
                the target values can reasonably be approximated by a sum of [0, 1] importance values.
            sparsity_factor: The coefficient for the sparsity regularization of the node importance
                tensor.
            concat_heads: Whether to concat the heads of the attention subnetwork. The default is True. In
                that case the output of each individual attention head is concatenated and the concatenated
                vector is then used as the input of the next attention layer's heads. If this is False, the
                vectors are average pooled instead.
            final_units: A list of ints where each element configures another dense layer in the MLP
                at the tail end of the network. The numeric value determines the number of the hidden units
                in that layer. Note that the final element in this list has to be the same as the dimension
                to be expected for the samples of the training dataset!
            final_dropout_rate: The dropout rate to be applied after *every* layer of the final MLP.
            final_activation: The activation to be applied at the very last layer of the MLP to produce the
                actual output of the network.
            final_pooling: The pooling method to be used during the global pooling phase in the network.
            regression_limits: A tuple where the first value is the lower limit for the expected value range
                of the regression task and teh second value the upper limit.
            regression_reference: A reference value which is inside the range of expected values (best if
                it was in the middle, but does not have to). Choosing different references will result
                in different explanations.
            return_importances: Whether the importance / explanation tensors should be returned as an output
                of the model. If this is True, the output of the model will be a 3-tuple:
                (output, node importances, edge importances), otherwise it is just the output itself
        """
        ks.models.Model.__init__(self, **kwargs)
        self.units = units
        self.activation = activation
        self.use_bias = use_bias
        self.dropout_rate = dropout_rate
        self.use_edge_features = use_edge_features
        self.importance_units = importance_units
        self.importance_channels = importance_channels
        self.importance_activation = importance_activation
        self.importance_dropout_rate = importance_dropout_rate
        self.importance_factor = importance_factor
        self.importance_multiplier = importance_multiplier
        self.importance_transformations = importance_transformations
        self.sparsity_factor = sparsity_factor
        self.concat_heads = concat_heads
        self.final_units = final_units
        self.final_dropout_rate = final_dropout_rate
        self.final_activation = final_activation
        self.final_pooling = final_pooling
        self.regression_limits = regression_limits
        self.regression_reference = regression_reference
        self.regression_bins = regression_bins
        self.return_importances = return_importances
        self.separate_explanation_step = separate_explanation_step
        self.use_graph_attributes = use_graph_attributes

        # ~ MAIN CONVOLUTIONAL / ATTENTION LAYERS
        self.attention_layers: t.List[GraphBaseLayer] = []
        for u in self.units:
            lay = MultiHeadGATV2Layer(
                units=u,
                num_heads=self.importance_channels,
                use_edge_features=self.use_edge_features,
                activation=self.activation,
                use_bias=self.use_bias,
                has_self_loops=True,
                concat_heads=self.concat_heads
            )
            self.attention_layers.append(lay)

        self.lay_dropout = DropoutEmbedding(rate=self.dropout_rate)
        self.lay_sparsity = ExplanationSparsityRegularization(factor=self.sparsity_factor)

        # ~ EDGE IMPORTANCES
        self.lay_act_importance = ActivationEmbedding(activation=self.importance_activation)
        self.lay_concat_alphas = LazyConcatenate(axis=-1)

        self.lay_pool_edges_in = PoolingLocalEdges(pooling_method='mean', pooling_index=0)
        self.lay_pool_edges_out = PoolingLocalEdges(pooling_method='mean', pooling_index=1)
        self.lay_average = LazyAverage()

        # ~ NODE IMPORTANCES
        self.node_importance_units = importance_units + [self.importance_channels]
        self.node_importance_acts = ['kgcnn>leaky_relu' for _ in importance_units] + ['linear']
        self.node_importance_layers = []
        for u, act in zip(self.node_importance_units, self.node_importance_acts):
            lay = DenseEmbedding(
                units=u,
                activation=act,
                use_bias=use_bias
            )
            self.node_importance_layers.append(lay)

        # ~ OUTPUT / MLP TAIL END
        self.lay_pool_out = PoolingNodes(pooling_method=self.final_pooling)
        self.lay_concat_out = LazyConcatenate(axis=-1)
        self.lay_final_dropout = DropoutEmbedding(rate=self.final_dropout_rate)

        self.final_acts = ['kgcnn>leaky_relu' for _ in self.final_units]
        self.final_acts[-1] = self.final_activation
        self.final_biases = [True for _ in self.final_units]
        self.final_biases[-1] = True
        self.final_layers = []
        for u, act, bias in zip(self.final_units, self.final_acts, self.final_biases):
            lay = DenseEmbedding(
                units=u,
                activation=act,
                use_bias=use_bias
            )
            self.final_layers.append(lay)

        # ~ EXPLANATION ONLY TRAIN STEP
        self.bce_loss = ks.losses.BinaryCrossentropy()
        self.compiled_classification_loss = compile_utils.LossesContainer(self.bce_loss)

        self.mse_loss = ks.losses.MeanSquaredError()
        self.mae_loss = ks.losses.MeanAbsoluteError()
        self.compiled_regression_loss = compile_utils.LossesContainer(mae)

        # If regression_limits have been supplied, we interprete this as the intent to perform explanation
        # co-training for a regression dataset.
        # So the content of this if condition makes sure to perform the necessary pre-processing steps
        # for this case.
        if self.regression_limits is not None:
            # first of all we do some simple assertions to handle some common error cases
            assert regression_reference is not None, ('You have to supply a "regression_reference" value for '
                                                      'explanation co-training!')

            # This is the first and simpler case for regression explanation co-training: In this case the
            # regression reference value is only a single value. In that case, there is only one target
            # value that is supposed to be regressed. The alternative would be that it is a list in which
            # case it would have to have as many elements as target values to be predicted.
            # However in this case we convert it into a list as well to be able to treat everything from
            # this point on as the multi-value case guaranteed.
            if isinstance(regression_reference, (int, float)):
                self.regression_reference = [regression_reference]

            num_references = len(self.regression_reference)
            num_limits = len(self.regression_limits)
            assert num_references * 2 == importance_channels, (
                f'for explanation co-training, the number of regression_references (currently {num_references}) has '
                f'to be exactly half the number of importance channels (currently {importance_channels})!'
            )
            assert num_references == final_units[-1], (
                f'For explanation co-training, the number of regression_references (currently {num_references}) has '
                f'to be exactly the same as the final unit count in the MLP tail end (currently {final_units[-1]})'
            )
            assert num_references == num_limits, (
                f'For explanation co-training, the number of regression_references (currently {num_references}) has '
                f'to be exactly the same as the number of regression_limits intervals (currently {num_limits})'
            )

    def get_config(self):
        config = super(Megan, self).get_config()
        config.update({
            "units": self.units,
            "activation": self.activation,
            "use_bias": self.use_bias,
            "dropout_rate": self.dropout_rate,
            "use_edge_features": self.use_edge_features,
            "importance_units": self.importance_units,
            "importance_channels": self.importance_channels,
            "importance_activation": self.importance_activation,
            "importance_dropout_rate": self.importance_dropout_rate,
            "importance_factor": self.importance_factor,
            "importance_multiplier": self.importance_multiplier,
            "sparsity_factor": self.sparsity_factor,
            "concat_heads": self.concat_heads,
            "final_units": self.final_units,
            "final_dropout_rate": self.final_dropout_rate,
            "final_activation": self.final_activation,
            "final_pooling": self.final_pooling,
            "regression_limits": self.regression_limits,
            "regression_reference": self.regression_reference,
            "return_importances": self.return_importances
        })

        return config
    @property
    def doing_regression(self) -> bool:
        return self.regression_limits is not None

    def call(self,
             inputs,
             training: bool = False,
             return_importances: bool = False,
             node_importances_mask: t.Optional[tf.RaggedTensor] = None,
             **kwargs):
        """
        Forward pass of the model.

        **Shape Explanations:** All shapes in brackets [] are ragged dimensions!

        - V: Num nodes in the graph
        - E: Num edges in the graph
        - N: Num feature values per node
        - M: NUm feature values per edge
        - H: Num feature values per graph
        - B: Num graphs in a batch
        - K: Num importance (explanation) channels configured in the constructor
        """

        # 17.11.2022
        # Added support for global graph attributes. If the corresponding flag is set in the constructor of
        # the model then it is expected that the input tuple consists of 4 elements instead of the usual
        # 3 elements, where the fourth element is the vector of the graph attributes.
        # We can't use these graph attributes right away, but later on we will simply append them to the
        # vector which enters the MLP tail end.

        # node_input: ([B], [V], N)
        # edge_input: ([B], [E], M)
        # edge_index_input: ([B], [E], 2)
        # graph_input: ([B], H)
        if self.use_graph_attributes:
            node_input, edge_input, edge_index_input, graph_input = inputs
        else:
            node_input, edge_input, edge_index_input = inputs
            graph_input = None

        # First of all we apply all the graph convolutional / attention layers. Each of those layers outputs
        # the attention logits alpha additional to the node embeddings. We collect all the attention logits
        # in a list so that we can later sum them all up.
        alphas = []
        x = node_input
        for lay in self.attention_layers:
            # x: ([batch], [N], F)
            # alpha: ([batch], [M], K, 1)
            x, alpha = lay([x, edge_input, edge_index_input])
            if training:
                x = self.lay_dropout(x, training=training)

            alphas.append(alpha)

        # We sum up all the individual layers attention logit tensors and the edge importances are directly
        # calculated by applying a sigmoid on that sum.
        alphas = self.lay_concat_alphas(alphas)
        edge_importances = tf.reduce_sum(alphas, axis=-1, keepdims=False)
        edge_importances = self.lay_act_importance(edge_importances)

        # Part of the final node importance tensor is actually the pooled edge importances, so that is what
        # we are doing here. The caveat here is that we assume undirected edges as two directed edges in
        # opposing direction. To now achieve a symmetric pooling of these edges we have to pool in both
        # directions and then use the average of both.
        pooled_edges_in = self.lay_pool_edges_in([node_input, edge_importances, edge_index_input])
        pooled_edges_out = self.lay_pool_edges_out([node_input, edge_importances, edge_index_input])
        pooled_edges = self.lay_average([pooled_edges_out, pooled_edges_in])

        node_importances_tilde = x
        for lay in self.node_importance_layers:
            node_importances_tilde = lay(node_importances_tilde)

        node_importances_tilde = self.lay_act_importance(node_importances_tilde)

        node_importances = node_importances_tilde * pooled_edges
        self.lay_sparsity(node_importances)

        # ~ Applying the node importance mask
        # "node_importances_mask" is supposed to be a ragged tensor of the exact same dimensions as the
        # node importances, containing binary values 0 or 1, which are then used as a multiplicative mask
        # to modify the actual node importances before the global pooling step.
        # The main use case of this feature is to completely mask out certain channels to see how that
        # the missing channels (linked to a certain explanation / interpretation) affect the outcome of
        # the MLP tail end.
        if node_importances_mask is not None:
            node_importances_mask = tf.cast(node_importances_mask, tf.float32)
            node_importances = node_importances * node_importances_mask

        # Here we apply the global pooling. It is important to note that we do K separate pooling operations
        # were each time we use the same node embeddings x but a different slice of the node importances as
        # the weights! We concatenate all the individual results in the end.
        outs = []
        n = self.final_units[-1]
        for k in range(self.importance_channels):
            node_importance_slice = tf.expand_dims(node_importances[:, :, k], axis=-1)
            masked_embeddings = x * node_importance_slice

            # 26.03.2023
            # Optionally, if given we apply an additional non-linear transformation in the form of an
            # additional layer on each of the masked node embeddings separately.
            if self.importance_transformations is not None:
                lay_transform = self.importance_transformations[k]
                masked_embeddings = lay_transform(masked_embeddings)

            out = self.lay_pool_out(masked_embeddings)
            # out = self.lay_pool_out(x[:, :, k*n:(k+1)*n] * node_importance_slice)

            outs.append(out)

        # out: ([B], N*K²)
        out = self.lay_concat_out(outs)

        # At this point, after the global pooling of the node embeddings, we can append the global graph
        # attributes, should those exist
        if self.use_graph_attributes:
            out = self.lay_concat_out([out, graph_input])

        # Now "out" is a graph embedding vector of known dimension so we can simply apply the normal dense
        # mlp to get the final output value.
        num_final_layers = len(self.final_layers)
        for c, lay in enumerate(self.final_layers):
            out = lay(out)
            if training and c < num_final_layers - 2:
                out = self.lay_final_dropout(out, training=training)

        if self.doing_regression:
            reference = tf.ones_like(out) * tf.constant(self.regression_reference, dtype=tf.float32)
            out = out + reference

        # Usually, the node and edge importance tensors would be direct outputs of the model as well, but
        # we need the option to just return the output alone to be compatible with the standard model
        # evaluation pipeline already implemented in the library.
        if self.return_importances or return_importances:
            return out, node_importances, edge_importances
        else:
            return out

    def regression_augmentation(self,
                                out_true):
        samples = []
        masks = []
        for i, (regression_reference, regression_limits) in enumerate(zip(self.regression_reference,
                                                                          self.regression_limits)):

            regression_width = abs(regression_limits[1] - regression_limits[0])
            values = tf.expand_dims(out_true[:, i], axis=-1)
            center_distances = tf.abs(values - regression_reference)
            center_distances = (center_distances * self.importance_multiplier) / (0.5 * regression_width)

            # So we need two things: a "samples" tensor and a "mask" tensor. We are going to use the samples
            # tensor as the actual ground truth which acts as the regression target during the explanation
            # train step. The binary values of the mask will determine at which positions a loss should
            # actually be calculated for both of the channels

            # The "lower" part is all the samples which have a target value below the reference value.
            lo_mask = tf.where(values < regression_reference, 1.0, 0.0)
            # The "higher" part is all the samples above reference
            hi_mask = tf.where(values > regression_reference, 1.0, 0.0)

            samples += [center_distances, center_distances]
            masks += [lo_mask, hi_mask]

        return (
            tf.concat(samples, axis=-1),
            tf.concat(masks, axis=-1)
        )

    def train_step_explanation(self, x, y,
                               update_weights: bool = True):

        if self.return_importances:
            out_true, _, _ = y
        else:
            out_true = y

        exp_loss = 0
        with tf.GradientTape() as tape:

            y_pred = self(x, training=True, return_importances=True)
            out_pred, ni_pred, ei_pred = y_pred

            # ~ explanation loss
            # First of all we need to assemble the approximated model output, which is simply calculated
            # by applying a global pooling operation on the corresponding slice of the node importances.
            # So for each slice (each importance channel) we get a single value, which we then
            # concatenate into an output vector with K dimensions.
            outs = []
            for k in range(self.importance_channels):
                node_importances_slice = tf.expand_dims(ni_pred[:, :, k], axis=-1)
                out = self.lay_pool_out(node_importances_slice)

                outs.append(out)

            # outs: ([batch], K)
            outs = self.lay_concat_out(outs)

            if self.doing_regression:
                out_true, mask = self.regression_augmentation(out_true)
                out_pred = outs
                exp_loss = self.importance_channels * self.compiled_regression_loss(out_true * mask,
                                                                                    out_pred * mask)

            else:
                # out_pred = ks.backend.sigmoid(outs)
                out_pred = shifted_sigmoid(outs, shift=self.importance_multiplier, multiplier=2)
                exp_loss = self.compiled_classification_loss(out_true, out_pred * out_true)

            exp_loss *= self.importance_factor

        # Compute gradients
        trainable_vars = self.trainable_variables
        exp_gradients = tape.gradient(exp_loss, trainable_vars)

        # Update weights
        if update_weights:
            self.optimizer.apply_gradients(zip(exp_gradients, trainable_vars))

        return {'exp_loss': exp_loss}

    def train_step(self, data):
        if len(data) == 3:
            x, y, sample_weight = data
        else:
            sample_weight = None
            x, y = data

        # The "train_step_explanation" method will execute the entire explanation train step once. This
        # mainly involves creating an approximate solution of the original regression / classification
        # problem using ONLY the node importances of the different channels and then performing one
        # complete weight update based on the corresponding loss.
        if self.importance_factor != 0 and self.separate_explanation_step:
            exp_metrics = self.train_step_explanation(x, y)
        else:
            exp_metrics = {}

        exp_loss = 0
        with tf.GradientTape() as tape:

            out_true, ni_true, ei_true = y
            out_pred, ni_pred, ei_pred = self(x, training=True, return_importances=True)
            loss = self.compiled_loss(
                [out_true, ni_true, ei_true],
                [out_pred, ni_pred, ei_pred],
                sample_weight=sample_weight,
                regularization_losses=self.losses,
            )

            if self.importance_factor != 0 and not self.separate_explanation_step:
                # ~ explanation loss
                # First of all we need to assemble the approximated model output, which is simply calculated
                # by applying a global pooling operation on the corresponding slice of the node importances.
                # So for each slice (each importance channel) we get a single value, which we then
                # concatenate into an output vector with K dimensions.
                outs = []
                for k in range(self.importance_channels):
                    node_importances_slice = tf.expand_dims(ni_pred[:, :, k], axis=-1)
                    out = self.lay_pool_out(node_importances_slice)

                    outs.append(out)

                # outs: ([batch], K)
                outs = self.lay_concat_out(outs)

                if self.doing_regression:
                    _out_true, mask = self.regression_augmentation(out_true)
                    _out_pred = outs
                    exp_loss = self.compiled_regression_loss(_out_true * mask,
                                                             _out_pred * mask)

                else:
                    # out_pred = ks.backend.sigmoid(outs)
                    #_out_pred = shifted_sigmoid(outs, multiplier=self.importance_multiplier)
                    _out_pred = shifted_sigmoid(
                        outs,
                        shift=self.importance_multiplier,
                        multiplier=1
                    ) * out_true
                    exp_loss = self.compiled_classification_loss(out_true, _out_pred)

                exp_loss *= self.importance_factor
                exp_metrics['exp_loss'] = exp_loss
                loss += exp_loss

        trainable_vars = self.trainable_variables
        gradients = tape.gradient(loss, trainable_vars)

        self.optimizer.apply_gradients(zip(gradients, trainable_vars))

        self.compiled_metrics.update_state(
            y,
            out_pred if not self.return_importances else [out_pred, ni_pred, ei_pred],
            sample_weight=sample_weight
        )

        # Return a dict mapping metric names to current value.
        # Note that it will include the loss (tracked in self.metrics).
        return {
            **{m.name: m.result() for m in self.metrics},
            **exp_metrics
        }


class GnesGrad(ks.models.Model):

    def __init__(self,
                 units: t.List[int],
                 batch_size: int,
                 num_outputs: int = 1,
                 pooling_method: str = 'mean'):
        super(GnesGrad, self).__init__()
        self.batch_size = batch_size

        self.conv_layers = []
        for k in units:
            #lay = AttentionHeadGAT(units=k)
            lay = GCN(units=k)
            self.conv_layers.append(lay)

        self.lay_pooling = PoolingNodes(pooling_method=pooling_method)
        self.lay_dense = DenseEmbedding(units=num_outputs)

    def call(self, inputs, batch_size=None):
        if batch_size is None:
            batch_size = self.batch_size

        node_input, edge_input, edge_index_input = inputs
        edge_input = tf.cast(edge_input, dtype=tf.float32)
        x = node_input

        with tf.GradientTape(persistent=True) as tape:
            tape.watch(node_input)
            tape.watch(edge_input)

            activations = []
            for lay in self.conv_layers:
                x = lay([x, edge_input, edge_index_input])
                activations.append(x)

            out = self.lay_pooling(x)
            out = self.lay_dense(out)

            out_list = [out[i, 0] for i in range(batch_size)]

        node_gradients = []
        edge_gradients = []
        for i in range(batch_size):
            out_value = out_list[i]

            node_gradient = tape.gradient(out_value, node_input)[i]
            node_gradients.append(node_gradient)

            edge_gradient = tape.gradient(out_value, edge_input)[i]
            edge_gradients.append(edge_gradient)

        node_gradients = tf.ragged.stack(node_gradients, axis=0)
        edge_gradients = tf.ragged.stack(edge_gradients, axis=0)

        node_importances = tf.reduce_mean(node_gradients, axis=-1, keepdims=True)
        node_importances = tf.concat([
            tf.where(node_importances > 0, node_importances, 0.),
            tf.where(node_importances < 0, -node_importances, 0.),
        ], axis=-1)

        edge_importances = tf.reduce_mean(edge_gradients, axis=-1, keepdims=True)
        edge_importances = tf.concat([
            tf.where(edge_importances < 0, -edge_importances, 0.),
            tf.where(edge_importances > 0, edge_importances, 0.),
        ], axis=-1)

        return out, node_importances, edge_importances


class GnesGradCam(ks.models.Model):

    def __init__(self,
                 units: t.List[int],
                 batch_size: int,
                 num_outputs: int = 1,
                 pooling_method: str = 'mean'):
        super(GnesGradCam, self).__init__()
        self.batch_size = batch_size

        self.conv_layers = []
        for k in units:
            #lay = AttentionHeadGAT(units=k)
            lay = GCN(units=k)
            self.conv_layers.append(lay)

        self.lay_pooling = PoolingNodes(pooling_method=pooling_method)
        self.lay_dense = DenseEmbedding(units=num_outputs)
        self.lay_concat = LazyConcatenate(axis=0)

    def call(self, inputs, batch_size=None):
        if batch_size is None:
            batch_size = self.batch_size

        node_input, edge_input, edge_index_input = inputs
        x = node_input

        with tf.GradientTape(persistent=True) as tape:
            tape.watch(node_input)
            tape.watch(edge_input)

            activations = []
            for lay in self.conv_layers:
                x = lay([x, edge_input, edge_index_input])
                activations.append(x)

            out = self.lay_pooling(x)
            out = self.lay_dense(out)

            out_list = [out[i, 0] for i in range(batch_size)]

        node_importances_list: t.List[t.List[tf.Tensor]] = []
        edge_importances_list: t.List[t.List[tf.Tensor]] = []
        for l, activation in enumerate(activations):
            node_gradients = []
            edge_gradients = []
            for i in range(batch_size):
                out_value = out_list[i]

                node_gradient = tape.gradient(out_value, activation)
                node_gradient = tf.expand_dims(node_gradient, axis=0)
                node_gradients.append(node_gradient)

            node_gradients = tf.concat(node_gradients, axis=0)
            node_gradients = tf.reduce_sum(node_gradients, axis=1)

            node_alpha = tf.reduce_mean(node_gradients, axis=1, keepdims=True)
            node_importances = activation * node_alpha
            node_importances = tf.reduce_sum(node_importances, axis=-1, keepdims=True)
            node_importances_list.append(node_importances)

        node_importances = tf.concat(node_importances_list, axis=-1)
        node_importances = tf.reduce_mean(node_importances, axis=-1, keepdims=True)
        node_importances = tf.concat([
            tf.where(node_importances < 0, -node_importances, 0.),
            tf.where(node_importances > 0, node_importances, 0.),
        ], axis=-1)

        edge_gradients = []
        for i in range(batch_size):
            out_value = out_list[i]

            edge_gradient = tape.gradient(out_value, edge_input)[i]
            edge_gradients.append(edge_gradient)

        edge_gradients = tf.ragged.stack(edge_gradients, axis=0)

        edge_importances = tf.reduce_mean(edge_gradients, axis=-1, keepdims=True)
        edge_importances = tf.concat([
            tf.where(edge_importances < 0, -edge_importances, 0.),
            tf.where(edge_importances > 0, edge_importances, 0.),
        ], axis=-1)

        return out, node_importances, edge_importances

    def train_step(self, data):
        # Unpack the data. Its structure depends on your model and
        # on what you pass to `fit()`.
        x, y = data

        with tf.GradientTape() as tape:
            y_pred = self(x, training=True)  # Forward pass
            # Compute the loss value
            # (the loss function is configured in `compile()`)
            loss = self.compiled_loss(y, y_pred, regularization_losses=self.losses)

        # Compute gradients
        trainable_vars = self.trainable_variables
        gradients = tape.gradient(loss, trainable_vars)
        # Update weights
        self.optimizer.apply_gradients(zip(gradients, trainable_vars))
        # Update metrics (includes the metric that tracks the loss)
        self.compiled_metrics.update_state(y, y_pred)
        # Return a dict mapping metric names to current value
        return {m.name: m.result() for m in self.metrics}


class AbstractGradientModel(ks.models.Model):
    """
    This is an abstract base class for models which will provide their gradient information for further
    processing. This means that during the inference call, these kinds of models have to maintain their
    own gradient tape to record the full inference process and then use this to provide gradients of the
    input and intermediate node / edge embeddings with respect to the output.
    """

    def call_with_gradients(self, inputs, training=True):
        """
        **How the list of gradients should look like**

        Assuming output dimension H, number of nodes V, node features N and batch size B, as well as the
        number of convolutional layers L in the network.

        The list should consist of L+1 elements, where the first element is always the gradients with respect
        to the input and after that the gradients of the intermediate node/edge embeddings.

        Each element should be a ragged tensor with the dimension ([B], [V], N, H)

        :return: This has to be a tuple of four values:
            - The main output prediction "y" of the model
            - A list containing the gradients of the input as well as the intermediate node(!) embeddings
              w.r.t to the output "y".
            - A list containing the gradients of the input as well as the intermediate edge(!) embeddings
              w.r.t to the output "y".
            - The gradient tape used to create the previously mentioned gradients
        """
        raise NotImplementedError()

    def call(self, inputs, training=True, *args, **kwargs):
        y, node_gradients, edge_gradients, tape = self.call_with_gradients(inputs, training, *args, **kwargs)
        return y

    def calculate_gradient_info(self,
                                outs: t.List[t.List[tf.Tensor]],
                                activations: t.List[tf.Tensor],
                                tape: tf.GradientTape,
                                batch_size):
        """
        This method is a boilerplate implementation that will calculate the gradient info list, given a list
        of model outputs, a list of layer activation tensors, the gradient tape and the batch size.

        Assuming the following variables:
        - B: batch dimension
        - O: output dimension
        - L: number of convolutional layers / number of intermediate embeddings w.r.t which the gradients
            are to be calculated
        - V: Node or edge shape
        - F: Feature shape

        :param outs: This is supposed to be a list of lists of zero dimensional tensors. Generally, the
            output of a regression / classification network will be a vector of shape ([B], O). This list
            has to basically be this tensor only in list format. This list should be created by applying
            the corresponding indexing slices to the output vector to split it into the nested format of
            zero dimensional tensor objects. Note that this indexing operation has to be done INSIDE the
            gradient tape context, otherwise the calculation of the gradients will not work!
        :param activations: This is a list with L elements, where each element is an intermediate node /
            edge embedding vector from the call process of the network. In this method one gradient vector
            will be created for each of the outputs w.r.t to each of these embeddings. The resulting
            gradient tensors will have the same shape as the embeddings!
        :param tape: The GradientTape which was used during the call process.
        :param batch_size: The batch size B that was used to obtain the outputs and the embeddings. This has
            to be exactly correct!
        :return: A list of tuples with the length L. Each tuple contains two elements. The first one is the
            exact same embedding tensor that was passed to the method in the "activations" parameter.
            This tensor has the shape ([B], [V], F). The second element is the corresponding gradient
            tensor that was created for that very embedding with the shape ([B], [V], F, O). Note that the
            gradient has an additional dimension for the output shape. This is because gradients are
            created for every output.
        """
        gradient_info = []
        for activation in activations:
            gradients = []
            for b in range(batch_size):

                # In this inner loop here we manage the fact that the network may have multiple outputs
                # O != 1, so we iterate over every one of the outputs for the current element of the batch
                # add them to "gradient_stack" and then concat them in the end. The important part here is
                # that we want to stack them such that the resulting shape is ([B], [V], O, N) and not
                # ([B], [V], N, O) as you would intuitively do it by concat over the last dimension!
                gradient_stack = []
                for out in outs[b]:
                    # What we get out of this operation is the following shape ([B], [V], N) but technically
                    # the first dimension is pointless, because we are calculating the gradients of only one
                    # specific output value in the batch with respect to every element in the batch. Only
                    # one element in this dimension will be != zero. But actually we need to do it this way
                    # so that the ragged concat operation further down the line will work. If we already only
                    # select the appropriate element here, we get a serious issue with the stacking of ragged
                    # dimensions later.
                    gradient = tape.gradient(out, activation)
                    gradient = tf.expand_dims(gradient, axis=-2)
                    gradient_stack.append(gradient)

                # gradient_stack will be a vector of shape ([B], [V], O, N, 1). The last dimension we have
                # expanded here because in the following we need to stack this over the batch dimension
                # again.
                gradient_stack = tf.concat(gradient_stack, axis=-2)
                gradient_stack = tf.expand_dims(gradient_stack, axis=-1)

                gradients.append(gradient_stack)

            # results in a tensor of shape ([B], [V], O, N, [B])
            gradients = tf.concat(gradients, axis=-1)
            # So this is the point AFTER the concat operation where we can savely get rid of the one
            # batch dimension we have uselessly kept before.
            # results in fully populated gradient vector of shape ([B], [V], N) exactly as we wanted
            gradients = tf.reduce_sum(gradients, axis=-1)
            gradient_info.append((activation, gradients))

        return gradient_info


class GcnGradientModel(AbstractGradientModel, GNNInterface):

    def __init__(self,
                 batch_size: int,
                 units: t.List[int],
                 final_units: t.List[int],
                 layer_cb: t.Callable = lambda units: GCN(units=units),
                 pooling_method: str = 'mean',
                 final_activation: str = 'linear'):
        AbstractGradientModel.__init__(self)
        GNNInterface.__init__(self)

        self.batch_size = batch_size
        self.units = units
        self.num_outputs = final_units[-1]

        self.conv_layers = []
        for k in self.units:
            lay = layer_cb(k)
            self.conv_layers.append(lay)

        self.lay_pooling = PoolingNodes(pooling_method=pooling_method)

        self.final_units = final_units
        self.final_acts = ['kgcnn>leaky_relu' for _ in final_units]
        self.final_acts[-1] = final_activation
        self.final_layers = []
        for k, act in zip(self.final_units, self.final_acts):
            lay = DenseEmbedding(units=k, activation=act)
            self.final_layers.append(lay)

    def call_with_gradients(self, inputs, training=True, create_gradients=True, batch_size=None):
        if batch_size is None:
            batch_size = self.batch_size

        node_input, edge_input, edge_index_input = inputs
        x = node_input

        with tf.GradientTape(persistent=True) as tape:
            tape.watch(node_input)
            tape.watch(edge_input)

            edge_activations = [edge_input]
            node_activations = [x]
            for lay in self.conv_layers:
                x = lay([x, edge_input, edge_index_input])
                node_activations.append(x)

            out = self.lay_pooling(x)

            for lay in self.final_layers:
                out = lay(out)

            outs = [out[b, :] for b in range(batch_size)]
            outs_multi = [[out[b, o] for o in range(self.num_outputs)] for b in range(batch_size)]

        if create_gradients:
            edge_gradient_info = self.calculate_gradient_info(outs_multi, edge_activations, tape, batch_size)
            node_gradient_info = self.calculate_gradient_info(outs_multi, node_activations, tape, batch_size)
        else:
            node_gradient_info = []
            edge_gradient_info = []

        return out, node_gradient_info, edge_gradient_info, tape

    def call(self,
             inputs,
             training=True,
             batch_size=None,
             create_gradients=True,
             return_gradients=False):
        y, node_info, edge_info, tape = self.call_with_gradients(
            inputs,
            training=training,
            batch_size=batch_size,
            create_gradients=create_gradients
        )

        if return_gradients:
            return y, node_info, edge_info
        else:
            return y


def grad_importances(gradient_info: t.List[t.Tuple[tf.Tensor, tf.Tensor]],
                     use_relu: bool = False,
                     use_absolute: bool = False,
                     keepdims: bool = False):
    # shapes as we get them:
    # gradients: ([B], [V], O, F)
    # activation: ([B], [V], F)

    activation, gradients = gradient_info[0]
    importances = tf.reduce_sum(
        # We need to expand the activations here in the second last dimension to match the additional
        # output shape O of the gradients.
        gradients * tf.expand_dims(activation, axis=-2),
        axis=-1,
        keepdims=keepdims
    )
    if use_relu:
        importances = ks.backend.relu(importances)
    if use_absolute:
        importances = tf.abs(importances)

    return importances


def grad_cam_importances(gradient_info: t.List[t.Tuple[tf.Tensor, tf.Tensor]],
                         use_average: bool = False,
                         use_relu: bool = False,
                         use_absolute: bool = False,
                         averaging_depth: int = 2,
                         keepdims: bool = False):
    importances = []
    for activation, gradients in gradient_info:
        # shapes as we get them:
        # gradients: ([B], [V], O, F)
        # activation: ([B], [V], F)

        alpha = tf.reduce_mean(gradients, axis=1, keepdims=True)
        local_importances = tf.reduce_sum(
            # We need to expand the activations here in the second last dimension to match the additional
            # output shape O of the gradients.
            alpha * tf.expand_dims(activation, axis=-2),
            axis=-1,
            keepdims=keepdims
        )

        if use_relu:
            local_importances = ks.backend.relu(local_importances)
        if use_absolute:
            local_importances = tf.abs(local_importances)

        importances.append(local_importances)

    if use_average:
        importances = tf.concat(importances[-averaging_depth:], axis=-1)
    else:
        importances = importances[-1]

    importances = tf.reduce_mean(importances, axis=-1, keepdims=True)
    return importances


def gnnx_importances(model,
                     x,
                     y,
                     epochs: int = 100,
                     learning_rate: float = 0.01,
                     node_sparsity_factor: float = 1.0,
                     edge_sparsity_factor: float = 1.0,
                     model_kwargs: dict = {},
                     logger: logging.Logger = NULL_LOGGER,
                     log_step: int = 10):
    logger.info('creating explanations with gnn explainer')
    start_time = time.time()

    optimizer = ks.optimizers.Nadam(learning_rate=learning_rate)

    node_input = x[0]
    edge_input = x[1]
    edge_index_input = x[2]

    node_mask_ragged = tf.reduce_mean(tf.ones_like(node_input), axis=-1, keepdims=True)
    node_mask_variables = tf.Variable(node_mask_ragged.flat_values, trainable=True, dtype=tf.float64)

    edge_mask_ragged = tf.reduce_mean(tf.ones_like(edge_input), axis=-1, keepdims=True)
    edge_mask_variables = tf.Variable(edge_mask_ragged.flat_values, trainable=True, dtype=tf.float64)

    for epoch in range(epochs):

        with tf.GradientTape() as tape:
            node_mask = tf.RaggedTensor.from_nested_row_splits(
                node_mask_variables,
                nested_row_splits=node_mask_ragged.nested_row_splits
            )
            node_masked = node_input * node_mask

            edge_mask = tf.RaggedTensor.from_nested_row_splits(
                edge_mask_variables,
                nested_row_splits=edge_mask_ragged.nested_row_splits
            )
            edge_masked = edge_input * edge_mask

            out = model([
                node_masked,
                edge_masked,
                edge_index_input,
                *x[3:]
            ], **model_kwargs)

            loss = tf.cast(tf.reduce_mean(tf.square(y - out)), dtype=tf.float64)
            loss += node_sparsity_factor * tf.reduce_mean(tf.abs(node_mask))
            loss += edge_sparsity_factor * tf.reduce_mean(tf.abs(edge_mask))

        trainable_vars = [node_mask_variables, edge_mask_variables]
        gradients = tape.gradient(loss, trainable_vars)
        optimizer.apply_gradients(zip(gradients, trainable_vars))

        if epoch % log_step == 0:
            logger.info(f' ({epoch}/{epochs})'
                        f' - loss: {loss}'
                        f' - elapsed time: {time.time() - start_time:.2f}')

    return (
        node_mask,
        edge_mask
    )


class GnesGradientModel(AbstractGradientModel):

    def __init__(self,
                 units: t.List[int],
                 batch_size: int,
                 importance_func: t.Callable,
                 layer_cb: t.Callable = lambda units: GCN(units=units, activation='kgcnn>leaky_relu'),
                 sparsity_factor: float = 0.1,
                 final_units: t.List[int] = [1],
                 final_activation: str = 'linear',
                 pooling_method: str = 'sum'):
        super(GnesGradientModel, self).__init__()
        self.importance_func = importance_func
        self.batch_size = batch_size
        self.num_outputs = final_units[-1]

        self.conv_layers = []
        for k in units:
            lay = layer_cb(k)
            self.conv_layers.append(lay)

        # ~ global pooling
        self.lay_pooling = PoolingNodes(pooling_method=pooling_method)

        # ~ mlp tail-end
        self.final_units = final_units
        self.final_acts = ['kgcnn>leaky_relu' for _ in self.final_units]
        self.final_acts[-1] = final_activation
        self.final_layers = []
        for k, act in zip(self.final_units, self.final_acts):
            lay = DenseEmbedding(
                units=k,
                activation=act
            )
            self.final_layers.append(lay)

        # ~ regularization
        self.lay_sparsity = ExplanationSparsityRegularization(factor=sparsity_factor)

    def call_with_gradients(self, inputs, training=True, batch_size=None, create_gradients=True):
        if batch_size is None:
            batch_size = self.batch_size

        node_input, edge_input, edge_index_input = inputs
        x = node_input

        with tf.GradientTape(persistent=True) as tape:
            tape.watch(node_input)
            tape.watch(edge_input)

            # ~ convolutional layers
            edge_activations = [edge_input]
            node_activations = [node_input]
            for lay in self.conv_layers:
                x = lay([x, edge_input, edge_index_input])
                node_activations.append(x)

            # ~ global pooling
            out = self.lay_pooling(x)

            # ~ mlp tail end
            for lay in self.final_layers:
                out = lay(out)

            outs = [[out[b, o] for o in range(self.num_outputs)] for b in range(batch_size)]

        if create_gradients:
            node_gradient_info = self.calculate_gradient_info(outs, node_activations, tape, batch_size)
            edge_gradient_info = self.calculate_gradient_info(outs, edge_activations, tape, batch_size)

            node_importances = self.importance_func(node_gradient_info)
            edge_importances = self.importance_func(edge_gradient_info)

            self.lay_sparsity(node_importances)
            self.lay_sparsity(edge_importances)

            return out, node_importances, edge_importances

        else:
            return out, None, None

    def call(self, inputs, *args, **kwargs):
        return self.call_with_gradients(inputs, *args, **kwargs)




"""
Trains multi-channel MEGAN models on the synthetic graph regression dataset RbMotifs. This datasets consits
of randomly generated colored graphs, where the target value of each graph is determined by the special
subgraph motifs contained in it. The MEGAN models are trained with two explanation channels: One to
represent the positive evidence w.r.t. to a reference value and the other for the negative evidence.
The experiment compares multiple hyperparameter configurations, mainly the impact of the "importance factor"
which is the weight assigned to the self-supervised explanation-only training step. The experiment is
repeated multiple independent times.
"""
import os
import warnings
warnings.filterwarnings("ignore")
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

import sys
import json
import time
import random
from typing import List, Callable, Dict, Optional
from pprint import pprint
from collections import Counter

import tensorflow as tf
import tensorflow.keras as ks
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
from matplotlib.backends.backend_pdf import PdfPages
from sklearn.metrics import roc_auc_score, r2_score
from imageio.v2 import imread
from pycomex.experiment import Experiment
from pycomex.util import Skippable

from graph_attention_student.util import DATASETS_FOLDER
from graph_attention_student.util import importance_absolute_similarity
from graph_attention_student.util import importance_canberra_similarity
from graph_attention_student.util import array_normalize
from graph_attention_student.util import binary_threshold
from graph_attention_student.util import render_latex, latex_table, latex_table_element_median
from graph_attention_student.data import load_eye_tracking_dataset
from graph_attention_student.data import process_graph_dataset
from graph_attention_student.models import MultiAttentionStudent
from graph_attention_student.training import NoLoss, ExplanationLoss
from graph_attention_student.training import LogProgressCallback
from graph_attention_student.visualization import plot_node_importances, plot_edge_importances

SHORT_DESCRIPTION = (
    'Trains multi-channel MEGAN model on the synthetic graph regression dataset RbMotifs. Compares the '
    'generated explanations to the known ground truth.'
)

# == META PARAMETERS ==
REPETITIONS = 50

# == DATASET PARAMETERS ==
DATASET_PATH = os.path.join(DATASETS_FOLDER, 'rb_dual_motifs')
TEST_RATIO = 0.1
NUM_EXAMPLES = 100
EXAMPLE_INDICES: Optional[list] = None

# == MODEL PARAMETERS ==
UNITS = [3, 3, 3]
IMPORTANCE_UNITS = []
SPARSITY_FACTOR = 0
IMPORTANCE_FACTOR = 0.0
IMPORTANCE_MULTIPLIER = 1
IMPORTANCE_SPARSITY = 0
NUM_CHANNELS = 2
FINAL_UNITS = [3]
DROPOUT_RATE = 0.0
FINAL_DROPOUT_RATE = 0.0
REGRESSION_LIMITS = [-4, 4]
REGRESSION_BINS = [
    [-5, -1.5],
    [1.5, 5]
]
USE_BIAS = True

MEGAN_SWEEP = {
    'megan_0': {
        'IMPORTANCE_FACTOR': 0.0,
        'SPARSITY_FACTOR': 1e-1,
        'IMPORTANCE_SUPERVISION': False,
    },
    'megan_1': {
        'IMPORTANCE_FACTOR': 0.1,
        'IMPORTANCE_MULTIPLIER': 10,
        'SPARSITY_FACTOR': 1e-1,
        'IMPORTANCE_SPARSITY': 0.,
        'IMPORTANCE_SUPERVISION': False,
    },
    'megan_2': {
        'IMPORTANCE_FACTOR': 1.0,
        'IMPORTANCE_MULTIPLIER': 10,
        'SPARSITY_FACTOR': 1e-1,
        'IMPORTANCE_SPARSITY': 0.,
        'IMPORTANCE_SUPERVISION': False,
    },
    'megan_3': {
        'IMPORTANCE_FACTOR': 0.0,
        'SPARSITY_FACTOR': 1e-1,
        'IMPORTANCE_SPARSITY': 0.,
        'IMPORTANCE_SUPERVISION': True,
    }
}

# == TRAINING PARAMETERS ==
LEARNING_RATE = 0.002
BATCH_SIZE = 512
EPOCHS = 250
LOG_STEP = 10
DEVICE = 'gpu:0'
METRIC_KEY = 'mean_squared_error'
IMPORTANCE_SUPERVISION = False

# == EVALUATION PARAMETERS ==
FIDELITY_SPARSITY = 0.15
BINARY_THRESHOLD = 0.5
LOG_STEP_EVAL = 20
CHANNEL_NAMES = {
    0: 'NEGATIVE',
    1: 'POSITIVE'
}
CHANNEL_DIRECTIONS = {
    0: -1,
    1: +1,
}

# == EXPERIMENT PARAMETERS ==
NAMESPACE = 'rb_motifs_multi'
BASE_PATH = os.getcwd()
DEBUG = True

with Skippable(), (e := Experiment(namespace=NAMESPACE, base_path=BASE_PATH, glob=globals())):

    e.info('Loading the dataset...')
    start_time = time.time()
    eye_tracking_dataset = load_eye_tracking_dataset(DATASET_PATH)
    dataset_size = len(eye_tracking_dataset)
    end_time = time.time()
    e.info(f'Loaded dataset with {dataset_size} elements in {end_time - start_time:.2f} seconds')
    e.status()

    e['metric_keys'] = []
    for rep in range(REPETITIONS):
        e.info(f'STARTING REPETITION ({rep+1}/{REPETITIONS})')

        dataset_indices = list(range(dataset_size))
        test_indices = random.sample(dataset_indices, k=int(TEST_RATIO * dataset_size))
        train_indices = [index for index in dataset_indices if index not in test_indices]
        e[f'test_indices/{rep}'] = test_indices
        e[f'train_indices/{rep}'] = train_indices
        e.info(f'Identified {len(train_indices)} elements for training and {len(test_indices)} for testing')

        example_indices = random.sample(test_indices, k=NUM_EXAMPLES)
        e[f'example_indices/{rep}'] = example_indices
        e.info(f'Identified {len(example_indices)} elements as examples')

        # "eye_tracking_dataset" at this point is only a list of dictionaries which contain the metadata info
        # for each respective elements. This now has to be turned into the appropriate tensors to be used by
        # the model.
        dataset = []
        for data in eye_tracking_dataset:
            g = data['metadata']['graph']

            # The target ground truth will be the float value associated with every graph
            value = data['metadata']['value']
            g['graph_labels'] = np.array(value)

            # The dataset already comes with ground truth importance annotations which are split into two
            # channels corresponding to "positive" and "negative" influences
            node_importances = np.array(g['multi_node_importances'])
            g['node_importances'] = node_importances

            edge_importances = np.array(g['multi_edge_importances'])
            g['edge_importances'] = edge_importances

            dataset.append(g)

        x_train, y_train, x_test, y_test = process_graph_dataset(dataset, test_indices)
        e.info('Pre-processed the dataset. Now ready for training')

        # ---------------------------------------------------------------------------------------------------
        # -- MULTI ATTENTION STUDENT ------------------------------------------------------------------------
        # ---------------------------------------------------------------------------------------------------
        for model_name, parameters in MEGAN_SWEEP.items():
            e.info(f'Model sweep: {model_name}')
            for parameter_key, parameter_value in parameters.items():
                globals()[parameter_key] = parameter_value
                e.info(f' * {parameter_key}: {eval(parameter_key)}')

            e.info('Training Multi Explanation Graph Attention Network (MEGAN) ...')
            ks.backend.clear_session()
            model: ks.models.Model = MultiAttentionStudent(
                units=UNITS,
                dropout_rate=DROPOUT_RATE,
                importance_factor=IMPORTANCE_FACTOR,
                importance_multiplier=IMPORTANCE_MULTIPLIER,
                importance_units=IMPORTANCE_UNITS,
                importance_channels=NUM_CHANNELS,
                importance_sparsity=IMPORTANCE_SPARSITY,
                use_bias=USE_BIAS,
                final_units=FINAL_UNITS,
                final_dropout_rate=FINAL_DROPOUT_RATE,
                regression_bins=REGRESSION_BINS,
                regression_limits=REGRESSION_LIMITS,
                sparsity_factor=SPARSITY_FACTOR
            )
            model.compile(
                loss=[
                    ks.losses.MeanSquaredError(),
                    NoLoss() if not IMPORTANCE_SUPERVISION else ExplanationLoss(),
                    NoLoss() if not IMPORTANCE_SUPERVISION else ExplanationLoss(),
                ],
                loss_weights=[
                    1,
                    0 if not IMPORTANCE_SUPERVISION else 1,
                    0 if not IMPORTANCE_SUPERVISION else 1,
                ],
                metrics=[
                    ks.metrics.MeanSquaredError(),
                    ks.metrics.MeanAbsoluteError(),
                ],
                optimizer=ks.optimizers.Adam(learning_rate=LEARNING_RATE),
                run_eagerly=False
            )
            with tf.device(DEVICE):
                history = model.fit(
                    x_train,
                    y_train,
                    batch_size=BATCH_SIZE,
                    epochs=EPOCHS,
                    validation_data=(x_test, y_test),
                    validation_freq=1,
                    callbacks=LogProgressCallback(
                        logger=e.logger,
                        epoch_step=LOG_STEP,
                        identifier=f'val_output_1_{METRIC_KEY}'
                    ),
                    verbose=0
                )
            e.info(f'model parameters: {model.count_params()}')
            e[f'histories/{model_name}/{rep}'] = history.history
            e[f'epochs/{model_name}/{rep}'] = list(range(EPOCHS))

            # ~ Making all the test set predictions and saving them
            ni_true_all = []
            ei_true_all = []

            ni_pred_all = []
            ei_pred_all = []

            e.info('Evaluating test set...')
            for c, index in enumerate(test_indices):
                data = eye_tracking_dataset[index]
                g = dataset[index]

                # First of all we need to query the model to make the actual prediction for the current element
                out_meg, ni_meg, ei_meg = model.predict_single((
                    g['node_attributes'],
                    g['edge_attributes'],
                    g['edge_indices']
                ))
                out_true = g['graph_labels'].tolist()
                e[f'out/true/{rep}/{index}'] = out_true
                e[f'out/{model_name}/{rep}/{index}'] = out_meg
                e[f'absolute_error/{model_name}/{rep}/{index}'] = np.abs(out_meg - out_true)
                e[f'squared_error/{model_name}/{rep}/{index}'] = np.square(out_meg - out_true)
                ni_true_all += np.array(g['node_importances']).flatten().tolist()
                ei_true_all += np.array(g['edge_importances']).flatten().tolist()
                # The first transformation we apply to the importances vectors is to apply a normalization on
                # them such that they use the entire value range [0, 1]
                ni_meg = array_normalize(ni_meg)
                ei_meg = array_normalize(ei_meg)
                # ni_mas = np.array(ni_mas)
                # ei_mas = np.array(ei_mas)
                e[f'ni/{model_name}/{rep}/{index}'] = ni_meg.tolist()
                e[f'ei/{model_name}/{rep}/{index}'] = ei_meg.tolist()
                ni_pred_all += ni_meg.flatten().tolist()
                ei_pred_all += ei_meg.flatten().tolist()
                # Then with these relatively raw importances we calculate the cont. similarity value (basically
                # something like a distance metric with a normalized & inverted scale)
                node_sim_abs_meg = importance_absolute_similarity(g['node_importances'], ni_meg)
                edge_sim_abs_meg = importance_absolute_similarity(g['edge_importances'], ei_meg)
                e[f'node_sim_abs/{model_name}/{rep}/{index}'] = node_sim_abs_meg
                e[f'edge_sim_abs/{model_name}/{rep}/{index}'] = edge_sim_abs_meg
                node_sim_can_meg = importance_canberra_similarity(g['node_importances'], ni_meg)
                edge_sim_can_meg = importance_canberra_similarity(g['edge_importances'], ei_meg)
                e[f'node_sim_can/{model_name}/{rep}/{index}'] = node_sim_can_meg
                e[f'edge_sim_can/{model_name}/{rep}/{index}'] = edge_sim_can_meg
                # Then we calculate the natural sparsity percentage of the importances. We do this by first
                # applying a simple threshold to turn the soft importance masks into hard importance masks
                # and then the sparsity is simply the mean of the hard mask.
                e[f'node_sparsity/{model_name}/{rep}/{index}'] = np.mean(binary_threshold(ni_meg, BINARY_THRESHOLD))
                e[f'edge_sparsity/{model_name}/{rep}/{index}'] = np.mean(binary_threshold(ei_meg, BINARY_THRESHOLD))
                # This part here also turns the soft masks into hard masks, but by turning the top K values
                # into 1s and the other values to 0s (basically dynamic threshold). We do this because that
                # is the fairer starting point for the auroc & fidelity calculation
                node_mask_count = int(FIDELITY_SPARSITY * len(g['node_indices']))
                edge_mask_count = int(FIDELITY_SPARSITY * len(g['edge_indices']))
                # ni_mas[:, 0] = binary_threshold_k(ni_mas[:, 0], k=node_mask_count).tolist()
                # ei_mas[:, 0] = binary_threshold_k(ei_mas[:, 0], k=edge_mask_count).tolist()

                # ~ IMPORTANCE MASKING FIDELITY
                fidelity_im = 0
                for k in range(NUM_CHANNELS):
                    _ni_mod = np.copy(ni_meg)
                    _ni_mod[:, k] = 0
                    _out_mod, _, _ = model.predict_single((
                        g['node_attributes'],
                        g['edge_attributes'],
                        g['edge_indices']
                    ), external_node_importances=_ni_mod)
                    fidelity_im += - CHANNEL_DIRECTIONS[k] * (_out_mod - out_meg)
                    e[f'out_mod_im/{model_name}/{rep}/{index}/{k}'] = _out_mod

                e[f'fidelity_im/{model_name}/{rep}/{index}'] = fidelity_im

                if c % LOG_STEP_EVAL == 0:
                    e.info(f' * {model_name.upper()} ({c}) '
                           f' - node sim abs: {node_sim_abs_meg:.2f}'
                           f' - edge sim abs: {edge_sim_abs_meg:.2f}'
                           f' - y_pred: {out_meg:.2f}'
                           f' - y_true: {e[f"out/true/{rep}/{index}"]:.2f}'
                           f' - fidelity_imp: {fidelity_im:.2f}')

            # ~ Calculating overarching metrics
            ni_auroc = roc_auc_score(ni_true_all, ni_pred_all)
            ei_auroc = roc_auc_score(ei_true_all, ei_pred_all)
            e[f'ni_auroc/{model_name}/{rep}'] = ni_auroc
            e[f'ei_auroc/{model_name}/{rep}'] = ei_auroc

            r2 = r2_score(
                list(e[f'out/true/{rep}'].values()),
                list(e[f'out/{model_name}/{rep}'].values())
            )
            e[f'r2/{model_name}/{rep}'] = r2

            e.info(f' * {model_name.upper()} TOTAL '
                   f' - r2: {r2:.2f}'
                   f' - node auroc: {ni_auroc:.2f}'
                   f' - edge auroc: {ei_auroc:.2f}')

        # ~ Drawing the examples
        e.info('Drawing examples...')
        examples_path = os.path.join(e.path, f'{rep}_examples.pdf')
        ncols = 1 + len(MEGAN_SWEEP)
        with PdfPages(examples_path) as pdf:
            for index in example_indices:
                g = dataset[index]
                data = eye_tracking_dataset[index]
                g['node_coordinates'] = np.array(g['node_coordinates'])
                image = np.asarray(imread(data['image_path']))

                fig, rows = plt.subplots(ncols=ncols, nrows=NUM_CHANNELS, figsize=(ncols * 8, 2 * 8),
                                         squeeze=False)
                fig.suptitle(f'Rep {rep} - Element {index}\n'
                             f'Ground Truth: {e[f"out/true/{rep}/{index}"]:.3f}')

                for row_index in range(NUM_CHANNELS):
                    column_index = 0

                    # -- GROUND TRUTH --
                    ax_gt = rows[row_index][column_index]
                    ax_gt.set_title('Ground Truth\n\n')
                    ax_gt.set_ylabel(f'Channel {row_index}: {CHANNEL_NAMES[row_index]}')
                    ax_gt.imshow(image, extent=(0, image.shape[0], 0, image.shape[1]))
                    node_importances = np.array(g['node_importances'])
                    plot_node_importances(
                        g=g,
                        ax=ax_gt,
                        vmax=np.max(node_importances[:, row_index]),
                        node_importances=node_importances[:, row_index],
                        node_coordinates=g['node_coordinates']
                    )
                    edge_importances = np.array(g['edge_importances'])
                    plot_edge_importances(
                        g=g,
                        ax=ax_gt,
                        vmax=np.max(edge_importances[:, row_index]),
                        edge_importances=edge_importances[:, row_index],
                        node_coordinates=g['node_coordinates']
                    )
                    column_index += 1

                    # -- MULTI ATTENTION STUDENT --
                    for model_name in MEGAN_SWEEP.keys():
                        ax_mas = rows[row_index][column_index]
                        if row_index == 0:
                            ax_mas.set_title(
                                f'Model "{model_name}"\n'
                                f'Prediction: {e[f"out/{model_name}/{rep}/{index}"]:.2f}\n'
                                f'ni sim abs: {e[f"node_sim_abs/{model_name}/{rep}/{index}"]:.2f} - '
                                f'ei sim abs: {e[f"edge_sim_abs/{model_name}/{rep}/{index}"]:.2f}\n '
                                f'ni sim can: {e[f"node_sim_can/{model_name}/{rep}/{index}"]:.2f} - '
                                f'ei sim can: {e[f"edge_sim_can/{model_name}/{rep}/{index}"]:.2f}'
                            )

                        ax_mas.imshow(image, extent=(0, image.shape[0], 0, image.shape[1]))
                        ni_mas = np.array(e[f'ni/{model_name}/{rep}/{index}'])
                        plot_node_importances(
                            g=g,
                            ax=ax_mas,
                            vmax=1,
                            #vmax=np.max(ni_mas),
                            #vmax=np.max(ni_mas[:, row_index]),
                            node_importances=ni_mas[:, row_index],
                            node_coordinates=g['node_coordinates']
                        )
                        ei_mas = np.array(e[f'ei/{model_name}/{rep}/{index}'])
                        plot_edge_importances(
                            g=g,
                            ax=ax_mas,
                            vmax=1,
                            #vmax=np.max(ei_mas),
                            #vmax=np.max(ei_mas[:, row_index]),
                            edge_importances=ei_mas[:, row_index],
                            node_coordinates=g['node_coordinates']
                        )
                        column_index += 1

                pdf.savefig(fig)
                plt.close(fig)

        e.info(f'finished repetition {rep + 1}')
        e.status()


with Skippable(), e.analysis:

    NUM_WORST_BEST = 50

    # -- LOADING DATASET ------------------------------------------------------------------------------------
    start_time = time.time()
    eye_tracking_dataset = load_eye_tracking_dataset(DATASET_PATH)
    dataset_size = len(eye_tracking_dataset)
    dataset_indices = list(range(dataset_size))

    # "eye_tracking_dataset" at this point is only a list of dictionaries which contain the metadata info
    # for each respective elements. This now has to be turned into the appropriate tensors to be used by
    # the model.
    dataset = []
    for data in eye_tracking_dataset:
        g = data['metadata']['graph']

        # The target ground truth will be the float value associated with every graph
        value = data['metadata']['value']
        g['graph_labels'] = np.array(value)

        # The dataset already comes with ground truth importance annotations which are split into two
        # channels corresponding to "positive" and "negative" influences
        node_importances = np.array(g['multi_node_importances'])
        g['node_importances'] = node_importances

        edge_importances = np.array(g['multi_edge_importances'])
        g['edge_importances'] = edge_importances

        dataset.append(g)

    end_time = time.time()
    e.info(f'Loaded dataset with {dataset_size} elements in {end_time - start_time:.2f} seconds')
    y_trues = [float(g['graph_labels']) for g in dataset]

    # -- ADDITIONAL CALCULATIONS ----------------------------------------------------------------------------
    e.info('Post-hoc calculations...')
    for rep in range(REPETITIONS):
        e.info(f'REP {rep + 1}/{REPETITIONS}')
        test_indices = e[f'test_indices/{rep}']
        for model_name in MEGAN_SWEEP.keys():
            ni_true_all = []
            ei_true_all = []

            ni_pred_all = []
            ei_pred_all = []
            for index in test_indices:
                g = dataset[index]
                ni_true = np.array(g['node_importances'])
                ei_true = np.array(g['edge_importances'])
                ni_true_all += ni_true.flatten().tolist()
                ei_true_all += ei_true.flatten().tolist()

                ni_pred = np.array(e[f'ni/{model_name}/{rep}/{index}'])
                ei_pred = np.array(e[f'ei/{model_name}/{rep}/{index}'])
                ni_pred_all += ni_pred.flatten().tolist()
                ei_pred_all += ei_pred.flatten().tolist()

            ni_auc = roc_auc_score(ni_true_all, ni_pred_all)
            ei_auc = roc_auc_score(ei_true_all, ei_pred_all)

            e[f'ni_auroc/{model_name}/{rep}'] = ni_auc
            e[f'ei_auroc/{model_name}/{rep}'] = ei_auc

            r2 = r2_score(
                list(e[f'out/true/{rep}'].values()),
                list(e[f'out/{model_name}/{rep}'].values())
            )
            e[f'r2/{model_name}/{rep}'] = r2

            e.info(f' * {model_name:<8} '
                   f' - ni auroc: {ni_auc:.3f}'
                   f' - ei auroc: {ei_auc:.3f}'
                   f' - r2: {r2:.3f}')

    e.info('TOTAL:')
    for model_name in MEGAN_SWEEP.keys():
        ni_aucs = []
        ei_aucs = []
        r2s = []
        for rep in range(REPETITIONS):
            ni_aucs.append(e[f'ni_auroc/{model_name}/{rep}'])
            ei_aucs.append(e[f'ei_auroc/{model_name}/{rep}'])
            r2s.append(e[f'r2/{model_name}/{rep}'])

        e.info(f' * {model_name:<8} '
               f' - ni auroc: {np.mean(ni_aucs):.2f} ({np.std(ni_aucs):.2f})'
               f' - ei auroc: {np.mean(ei_aucs):.2f} ({np.std(ei_aucs):.2f})'
               f' - r2: {np.mean(r2s):.2f} ({np.std(r2s):.2f})')

    # -- SIMPLE STATISTICS ----------------------------------------------------------------------------------
    e.info('Printing simple statistics')
    metrics = ['absolute_error', 'squared_error',
               'node_sim_abs', 'node_sim_can', 'edge_sim_abs', 'edge_sim_can',
               'node_sparsity', 'edge_sparsity',
               'fidelity_im']
    model_names = list(e.parameters['MEGAN_SWEEP'].keys())

    for metric in metrics:
        e.info('')
        e.info(metric.upper())
        for model_name in model_names:
            metric_values = []
            for rep in range(REPETITIONS):
                for index in e[f'{metric}/{model_name}/{rep}'].keys():
                    value = e[f'{metric}/{model_name}/{rep}/{index}']
                    if value is not None:
                        metric_values.append(value)

            e.info(f'')
            e.info(f'* {model_name} / {metric}')
            e.info(f'  - min: {np.min(metric_values):.2f} - max: {np.max(metric_values):.2f}')
            e.info(f'  - mean: {np.mean(metric_values):.2f}: (std: {np.std(metric_values):.2f})')
            e.info(f'  - (0.25 - median - 0.75): ({np.quantile(metric_values, 0.25):.2f} - '
                   f'{np.median(metric_values):.2f} - {np.quantile(metric_values, 0.75):.2f})')

    # -- RENDERING LATEX TABLE ------------------------------------------------------------------------------
    column_names = [
        'Model',
        'MSE',
        '$R^2$',
        'Node AUROC',
        'Edge AUROC',
        'Sparsity',
        'Fidelity',
        'Fidelity Rand.'
    ]
    rows = []
    for model_name, parameters in MEGAN_SWEEP.items():
        row = []
        model_identifier = r'$\text{MEGAN}^2_{' + f'{parameters["IMPORTANCE_FACTOR"]:.2f}' + r'}$'
        row.append(model_identifier)
        row.append([e[f'squared_error/{model_name}/{rep}/{index}']
                    for rep in range(REPETITIONS) for index in e[f'test_indices/{rep}']])
        row.append([e[f'r2/{model_name}/{rep}']
                    for rep in range(REPETITIONS)])
        row.append([e[f'ni_auroc/{model_name}/{rep}']
                    for rep in range(REPETITIONS)])
        row.append([e[f'ei_auroc/{model_name}/{rep}']
                    for rep in range(REPETITIONS)])
        row.append([e[f'node_sparsity/{model_name}/{rep}/{index}']
                    for rep in range(REPETITIONS) for index in e[f'test_indices/{rep}']])
        row.append([e[f'fidelity_im/{model_name}/{rep}/{index}']
                    for rep in range(REPETITIONS) for index in e[f'test_indices/{rep}']])
        row.append('-')
        rows.append(row)

    content, table = latex_table(
        column_names=column_names,
        rows=rows,
        list_element_cb=latex_table_element_median
    )
    print(table)
    e.commit_raw('table.tex', table)
    pdf_path = os.path.join(e.path, 'table.pdf')
    render_latex({'content': table}, output_path=pdf_path)
    e.info('rendered latex table')

    # -- DRAWING EXAMPLES -----------------------------------------------------------------------------------
    e.info('')
    e.info('Drawing best and worst examples according to different criteria')


    def combined_absolute_similarity(rep: int, index: int, model_name: Optional[str] = 'megan_2'):
        model_values = {}
        for name in e['node_sim_abs'].keys():
            node_sim = e[f'node_sim_abs/{name}/{rep}/{index}']
            edge_sim = e[f'edge_sim_abs/{name}/{rep}/{index}']
            model_values[name] = node_sim + edge_sim

        if model_name is None:
            return np.max(model_values.values())

        else:
            return model_values[model_name]


    def combined_canberra_similarity(rep: int, index: int, model_name: Optional[str] = 'megan_2'):
        model_values = {}
        for name in e['node_sim_can'].keys():
            node_sim = e[f'node_sim_can/{name}/{rep}/{index}']
            edge_sim = e[f'edge_sim_can/{name}/{rep}/{index}']
            model_values[name] = node_sim + edge_sim

        if model_name is None:
            return np.max(model_values.values())

        else:
            return model_values[model_name]


    criteria = [
        (combined_absolute_similarity, ['node_sim_abs', 'edge_sim_abs']),
        (combined_canberra_similarity, ['node_sim_can', 'edge_sim_can'])
    ]
    for criterion, metrics in criteria:
        criterion_name = criterion.__name__
        criterion_description = criterion.__doc__
        e.info(f' * {criterion_name}')

        value_tuples = []
        for rep in range(REPETITIONS):
            for index in e[f'test_indices/{rep}']:
                value = criterion(rep, index)
                value_tuples.append(((rep, index), value))

        # Now we need to sort this list according to the values and based on that sorting we can simply
        # get the N first and last elements of that list which will be the best and worst elements
        value_tuples = sorted(value_tuples, key=lambda t: t[1])

        best_tuples = [value_tuples[i] for i in range(NUM_WORST_BEST)]
        worst_tuples = list(reversed([value_tuples[-(i + 1)] for i in range(NUM_WORST_BEST)]))

        # And then we can generate the pdf with the examples
        path = os.path.join(e.path, f'{criterion_name}__best_worst.pdf')
        ncols = 1 + len(MEGAN_SWEEP)
        with PdfPages(path) as pdf:

            for (rep, index), value in reversed(best_tuples + worst_tuples):
                g = dataset[index]
                data = eye_tracking_dataset[index]
                g['node_coordinates'] = np.array(g['node_coordinates'])
                image = np.asarray(imread(data['image_path']))

                fig, rows = plt.subplots(ncols=ncols, nrows=NUM_CHANNELS, figsize=(ncols * 8, 2 * 8),
                                         squeeze=False)
                fig.suptitle(f'Rep {rep} - Element {index}\n'
                             f'{criterion_name}: {value:.2f}')

                for row_index in range(NUM_CHANNELS):
                    column_index = 0

                    # -- GROUND TRUTH --
                    ax_gt = rows[row_index][column_index]
                    ax_gt.set_title('Ground Truth\n\n')
                    ax_gt.set_ylabel(f'Channel {row_index}: {CHANNEL_NAMES[row_index]}')
                    ax_gt.imshow(image, extent=(0, image.shape[0], 0, image.shape[1]))
                    node_importances = np.array(g['node_importances'])
                    plot_node_importances(
                        g=g,
                        ax=ax_gt,
                        vmax=np.max(node_importances[:, row_index]),
                        node_importances=node_importances[:, row_index],
                        node_coordinates=g['node_coordinates']
                    )
                    edge_importances = np.array(g['edge_importances'])
                    plot_edge_importances(
                        g=g,
                        ax=ax_gt,
                        vmax=np.max(edge_importances[:, row_index]),
                        edge_importances=edge_importances[:, row_index],
                        node_coordinates=g['node_coordinates']
                    )
                    column_index += 1

                    # -- MULTI ATTENTION STUDENT --
                    for model_name in MEGAN_SWEEP.keys():
                        ax_mas = rows[row_index][column_index]
                        if row_index == 0:
                            title_lines = [f'Model {model_name}']
                            for metric in metrics:
                                value = e[f'{metric}/{model_name}/{rep}/{index}']
                                title_lines.append(f'{metric}: {value:.2f}')
                            ax_mas.set_title('\n'.join(title_lines))

                        ax_mas.imshow(image, extent=(0, image.shape[0], 0, image.shape[1]))
                        ni_mas = np.array(e[f'ni/{model_name}/{rep}/{index}'])
                        plot_node_importances(
                            g=g,
                            ax=ax_mas,
                            vmax=np.max(ni_mas),
                            # vmax=np.max(ni_mas[:, row_index]),
                            node_importances=ni_mas[:, row_index],
                            node_coordinates=g['node_coordinates']
                        )
                        ei_mas = np.array(e[f'ei/{model_name}/{rep}/{index}'])
                        plot_edge_importances(
                            g=g,
                            ax=ax_mas,
                            vmax=np.max(ei_mas),
                            # vmax=np.max(ei_mas[:, row_index]),
                            edge_importances=ei_mas[:, row_index],
                            node_coordinates=g['node_coordinates']
                        )
                        column_index += 1

                pdf.savefig(fig)
                plt.close(fig)